from __future__ import annotations

import random
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pytest

from sdgx.metrics.pair_column.mi_sim import MISim

# 创建测试数据


def generate_random_time(start_date, end_date):
    start_datetime = datetime.strptime(start_date, "%Y-%m-%d")
    end_datetime = datetime.strptime(end_date, "%Y-%m-%d")

    random_time_delta = random.randint(0, int((end_datetime - start_datetime).total_seconds()))
    random_datetime = start_datetime + timedelta(seconds=random_time_delta)

    return random_datetime


@pytest.fixture
def test_data_time():
    start_date = "1900-01-01"
    end_date = "2023-12-31"
    df = pd.DataFrame(
        {
            "time_x": [generate_random_time(start_date, end_date) for _ in range(10)],
            "time_y": [generate_random_time(start_date, end_date) for _ in range(10)],
        }
    )
    return df


@pytest.fixture
def test_data_category():
    role_set = ["admin", "user", "guest"]
    df = pd.DataFrame(
        {
            "role1": [random.choice(role_set) for _ in range(10)],
            "role2": [random.choice(role_set) for _ in range(10)],
        }
    )
    return df


@pytest.fixture
def test_data_num():
    df = pd.DataFrame(
        {
            "feature_x": [random.random() for _ in range(10)],
            "feature_y": [random.random() for _ in range(10)],
        }
    )
    return df


@pytest.fixture
def mi_sim_instance():
    return MISim()


def test_MISim_discrete(test_data_category, mi_sim_instance):
    metadata = {"role1": "category", "role2": "category"}
    col_src = "role1"
    col_tar = "role2"
    result = mi_sim_instance.calculate(
        test_data_category[col_src], test_data_category[col_tar], metadata
    )
    result1 = mi_sim_instance.calculate(
        test_data_category[col_src], test_data_category[col_src], metadata
    )
    result2 = mi_sim_instance.calculate(
        test_data_category[col_tar], test_data_category[col_src], metadata
    )

    assert result >= 0
    assert result <= 1
    assert result1 == 1
    assert round(result2, 9) == round(result, 9)


def test_MISim_continuous(test_data_num, mi_sim_instance):
    metadata = {"feature_x": "numerical", "feature_y": "numerical"}
    col_src = "feature_x"
    col_tar = "feature_y"
    result = mi_sim_instance.calculate(test_data_num[col_src], test_data_num[col_tar], metadata)
    result1 = mi_sim_instance.calculate(test_data_num[col_src], test_data_num[col_src], metadata)
    result2 = mi_sim_instance.calculate(test_data_num[col_tar], test_data_num[col_src], metadata)

    assert result >= 0
    assert result <= 1
    assert np.isclose(result1, 1, atol=1e-9)
    assert np.isclose(result, result2, atol=1e-9)


def test_MISim_time(test_data_time, mi_sim_instance):
    metadata = {"time_x": "datetime", "time_y": "datetime"}
    col_src = "time_x"
    col_tar = "time_y"
    result = mi_sim_instance.calculate(test_data_time[col_src], test_data_time[col_tar], metadata)
    result1 = mi_sim_instance.calculate(test_data_time[col_src], test_data_time[col_src], metadata)
    result2 = mi_sim_instance.calculate(test_data_time[col_tar], test_data_time[col_src], metadata)

    assert result >= 0
    assert result <= 1
    assert np.isclose(result1, 1, atol=1e-9)
    assert np.isclose(result, result2, atol=1e-9)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from typing import Any, Generator

import pandas as pd


class DataExporter:
    def write(self, dst: Any, data: pd.DataFrame | Generator[pd.DataFrame, None, None]) -> None:
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Dict, Set

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.transformers.base import Transformer
from sdgx.utils import logger


class NumericValueTransformer(Transformer):
    """
    A transformer class for numeric data.

    This class is used to transform numeric data by scaling it using the StandardScaler from sklearn.

    Attributes:
        standard_scale (bool): A flag indicating whether to scale the data using StandardScaler.
        int_columns (Set): A set of column names that are of integer type.
        float_columns (Set): A set of column names that are of float type.
        scalers (Dict): A dictionary of scalers for each numeric column.
    """

    standard_scale: bool = True
    """
    A flag indicating whether to scale the data using StandardScaler.
    If True, the data will be scaled using StandardScaler.
    If False, the data will not be scaled.
    """

    int_columns: Set
    """
    A set of column names that are of integer type.
    These columns will be considered for scaling if `standard_scale` is True.
    """

    float_columns: Set
    """
    A set of column names that are of float type.
    These columns will be considered for scaling if `standard_scale` is True.
    """

    scalers: Dict
    """
    A dictionary of scalers for each numeric column.
    The keys are the column names and the values are the corresponding scalers.
    """

    def __init__(self):
        self.int_columns = set()
        self.float_columns = set()
        self.scalers = {}

    def fit(
        self,
        metadata: Metadata | None = None,
        tabular_data: DataLoader | pd.DataFrame = None,
        **kwargs: dict[str, Any],
    ):
        """
        The fit method.

        Data columns of int and float types need to be recorded here (Get data from metadata).
        """

        # get exact final data type from metadata
        # int columns
        for each_col in metadata.int_columns:
            if each_col not in metadata.column_list:
                continue
            if metadata.get_column_data_type(each_col) == "int":
                self.int_columns.add(each_col)
                continue
            if metadata.get_column_data_type(each_col) == "id":
                self.int_columns.add(each_col)

        # float columns
        for each_col in metadata.float_columns:
            if each_col not in metadata.column_list:
                continue
            if metadata.get_column_data_type(each_col) == "float":
                self.float_columns.add(each_col)

        if len(self.int_columns) == 0 and len(self.float_columns) == 0:
            logger.info("NumericValueTransformer Fitted (No numeric columns).")
            return

        # fit each columnxf
        for each_col in list(self.int_columns) + list(self.float_columns):
            self._fit_column(each_col, tabular_data[[each_col]])

        self.fitted = True
        logger.info("NumericValueTransformer Fitted.")

    def _fit_column(self, column_name: str, column_data: pd.DataFrame) -> np.ndarray:
        """
        Fit every numeric (include int and float) column in `_fit_column`.
        """

        if self.standard_scale:
            self._fit_column_scale(column_name, column_data)
            return

        return

    def _fit_column_scale(self, column_name: str, column_data: pd.DataFrame) -> np.ndarray:
        """
        Fit every numeric (include int and float) column using sklearn StandardScaler.
        """

        self.scalers[column_name] = StandardScaler()
        self.scalers[column_name].fit(column_data)

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        Convert method to handle missing values in the input data.
        """

        logger.info("Converting data using NumericValueTransformer...")

        if len(self.int_columns) == 0 and len(self.float_columns) == 0:
            logger.info("Converting data using NumericValueTransformer... Finished (No column).")
            return

        processed_data = raw_data.copy()

        for each_col in list(self.int_columns) + list(self.float_columns):
            # convert every column then change the column
            processed_col = self._covert_column(each_col, processed_data[[each_col]])
            processed_data[each_col] = processed_col

        logger.info("Converting data using NumericValueTransformer... Finished.")
        return processed_data

    def _covert_column(self, column_name: str, column_data: pd.DataFrame):
        """
        Convert every numeric (include int and float) column.
        """

        if self.standard_scale:
            return self._covert_column_scale(column_name=column_name, column_data=column_data)
        pass

    def _covert_column_scale(self, column_name: str, column_data: pd.DataFrame):
        """
        Convert every numeric (include int and float) column using sklearn StandardScaler.
        """

        scaled_data = self.scalers[column_name].transform(column_data)
        return scaled_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        Reverse convert method, convert generated data into processed data.
        """

        for each_col in list(self.int_columns) + list(self.float_columns):
            # reverse convert every column then change the column
            processed_col = self._reverse_convert_column(each_col, processed_data[[each_col]])
            processed_data[each_col] = processed_col

        logger.info("Data reverse-converted by NumericValueTransformer (No Action).")

        return processed_data

    def _reverse_convert_column(self, column_name: str, column_data: pd.DataFrame):
        """
        Reverse convert method for each column.
        """

        if self.standard_scale:
            return self._reverse_convert_column_scale(
                column_name=column_name, column_data=column_data
            )
        return

    def _reverse_convert_column_scale(self, column_name: str, column_data: pd.DataFrame):
        """
        Reverse convert method for input column using scale method.
        """

        reverse_converted_data = self.scalers[column_name].inverse_transform(column_data)
        return reverse_converted_data

    pass


@hookimpl
def register(manager):
    manager.register("NumericValueTransformer", NumericValueTransformer)



"""Transformer for data that contains Null values."""

import logging

import numpy as np
import pandas as pd

LOGGER = logging.getLogger(__name__)


class NullTransformer:
    """Transformer for data that contains Null values.

    Args:
        missing_value_replacement (object or None):
            Indicate what to do with the null values. If an integer, float or string is given,
            replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are given,
            replace them with the corresponding aggregation (``'mean'`` only works for numerical
            values). If ``None`` is given, do not replace them. Defaults to ``None``.
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
    """

    nulls = None
    _model_missing_values = None
    _missing_value_replacement = None
    _null_percentage = None

    def __init__(self, missing_value_replacement=None, model_missing_values=False):
        self._missing_value_replacement = missing_value_replacement
        self._model_missing_values = model_missing_values

    def models_missing_values(self):
        """Indicate whether this transformer creates a null column on transform.

        Returns:
            bool:
                Whether a null column is created on transform.
        """
        return self._model_missing_values

    def _get_missing_value_replacement(self, data):
        """Get the fill value to use for the given data.

        Args:
            data (pd.Series):
                The data that is being transformed.

        Return:
            object:
                The fill value that needs to be used.
        """
        if self._missing_value_replacement is None:
            return None

        if self._missing_value_replacement == "mean":
            return data.mean()

        if self._missing_value_replacement == "mode":
            return data.mode(dropna=True)[0]

        return self._missing_value_replacement

    def fit(self, data):
        """Fit the transformer to the data.

        Evaluate if the transformer has to create the null column or not.

        Args:
            data (pandas.Series):
                Data to transform.
        """
        null_values = data.isna().to_numpy()
        self.nulls = null_values.any()

        self._missing_value_replacement = self._get_missing_value_replacement(data)
        if not self.nulls and self._model_missing_values:
            self._model_missing_values = False
            guidance_message = (
                f"Guidance: There are no missing values in column {data.name}. "
                "Extra column not created."
            )
            LOGGER.info(guidance_message)

        if not self._model_missing_values:
            self._null_percentage = null_values.sum() / len(data)

    def transform(self, data):
        """Replace null values with the indicated ``missing_value_replacement``.

        If required, create the null indicator column.

        Args:
            data (pandas.Series or numpy.ndarray):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        isna = data.isna()
        if isna.any() and self._missing_value_replacement is not None:
            data = data.fillna(self._missing_value_replacement)

        if self._model_missing_values:
            return pd.concat([data, isna.astype(np.float64)], axis=1).to_numpy()

        return data.to_numpy()

    def reverse_transform(self, data):
        """Restore null values to the data.

        If a null indicator column was created during fit, use it as a reference.
        Otherwise, randomly replace values with ``np.nan``. The percentage of values
        that will be replaced is the percentage of null values seen in the fitted data.

        Args:
            data (numpy.ndarray):
                Data to transform.

        Returns:
            pandas.Series
        """
        data = data.copy()
        if self._model_missing_values:
            if self.nulls:
                isna = data[:, 1] > 0.5

            data = data[:, 0]

        elif self.nulls:
            isna = np.random.random((len(data),)) < self._null_percentage

        data = pd.Series(data)

        if self.nulls and isna.any():
            data.loc[isna] = np.nan

        return data



from __future__ import annotations

import random

import numpy as np
import pandas as pd
import pytest

from sdgx.metrics.column.jsd import JSD


# 创建测试数据
@pytest.fixture
def dummy_data(dummy_single_table_path):
    yield pd.read_csv(dummy_single_table_path)


@pytest.fixture
def test_data():
    role_set = ["admin", "user", "guest"]
    df = pd.DataFrame(
        {
            "role": [random.choice(role_set) for _ in range(10)],
            "feature_x": [random.random() for _ in range(10)],
        }
    )
    return df


@pytest.fixture
def jsd_instance():
    return JSD()


def test_jsd_discrete(dummy_data, test_data, jsd_instance):
    cols = ["role"]
    result = jsd_instance.calculate(dummy_data, test_data, cols, discrete=True)
    result1 = jsd_instance.calculate(dummy_data, dummy_data, cols, discrete=True)
    result2 = jsd_instance.calculate(test_data, dummy_data, cols, discrete=True)

    assert result >= 0
    assert result <= 1
    assert np.isclose(result1, 0, atol=1e-9)
    assert np.isclose(result, result2, atol=1e-9)


def test_jsd_continuous(dummy_data, test_data, jsd_instance):
    cols = ["feature_x"]
    result = jsd_instance.calculate(dummy_data, test_data, cols, discrete=False)
    result1 = jsd_instance.calculate(dummy_data, dummy_data, cols, discrete=False)

    assert result >= 0
    assert result <= 1
    assert np.isclose(result1, 0, atol=1e-9)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

import json
from collections import namedtuple
from pathlib import Path
from typing import Any, Iterable, List, Union

from pydantic import BaseModel

from sdgx.exceptions import RelationshipInitError

KeyTuple = namedtuple("KeyTuple", ["parent", "child"])


class Relationship(BaseModel):
    """Relationship between tables

    For parent table, we don't need define primary key here.
    The primary key is pre-defined in parent table's metadata.

    Child table's foreign key should be defined here.
    """

    version: str = "1.0"

    # table names
    parent_table: str

    child_table: str

    foreign_keys: List[KeyTuple]
    """
    foreign keys.

    If key is a tuple, the first element is parent column name and the second element is child column name
    """

    @classmethod
    def build(
        cls,
        parent_table: str,
        child_table: str,
        foreign_keys: Iterable[str | tuple[str, str] | KeyTuple],
        parent_metadata: Metadata | None = None,
        child_metadata: Metadata | None = None,
    ) -> "Relationship":
        """
        Build relationship from parent table, child table and foreign keys

        Args:
            parent_table (str): parent table
            parent_metadata : metadata of parent table
            child_table (str): child table
            child_metadata : metadata of child table
            foreign_keys (Iterable[str | tuple[str, str]]): foreign keys. If key is a tuple, the first element is parent column name and the second element is child column name
        """

        if not parent_table:
            raise RelationshipInitError("parent table cannot be empty")
        if not child_table:
            raise RelationshipInitError("child table cannot be empty")

        foreign_keys = [
            KeyTuple(key, key) if isinstance(key, str) else KeyTuple(*key) for key in foreign_keys
        ]

        if not foreign_keys:
            raise RelationshipInitError("foreign keys cannot be empty")
        if parent_table == child_table:
            raise RelationshipInitError("child table and parent table cannot be the same")
        if parent_metadata and child_metadata:
            for key in foreign_keys:
                if type(parent_metadata) is not dict:
                    if key[0] not in parent_metadata.id_columns:
                        raise RelationshipInitError("type of foreign key in parent table is not id")
                    if key[1] not in child_metadata.id_columns:
                        raise RelationshipInitError("type of foreign key in child table is not id")
                else:  # if load from json file, Metadata is a dict
                    if key[0] not in parent_metadata["id_columns"]:
                        raise RelationshipInitError("type of foreign key in parent table is not id")
                    if key[1] not in child_metadata["id_columns"]:
                        raise RelationshipInitError("type of foreign key in child table is not id")
        return cls(
            parent_table=parent_table,
            child_table=child_table,
            foreign_keys=foreign_keys,
        )

    def _dump_json(self):
        return self.model_dump_json()

    def save(self, path: str | Path):
        """
        Save relationship to json file.
        """

        with path.open("w") as f:
            f.write(self._dump_json())

    @classmethod
    def load(cls, path: str | Path) -> "Relationship":
        """
        Load relationship from json file.
        """

        path = Path(path).expanduser().resolve()
        fields = json.load(path.open("r"))
        # print(fields)
        version = fields.pop("version", None)
        if version:
            cls.upgrade(version, fields)

        return Relationship.build(**fields)

    @classmethod
    def upgrade(cls, old_version: str, fields: dict[str, Any]) -> None:
        pass



"""CLI."""

import argparse

from sdgx.models.components.sdv_ctgan.data import read_csv, read_tsv, write_tsv
from sdgx.models.components.sdv_ctgan.synthesizers.ctgan import CTGAN


def _parse_args():
    parser = argparse.ArgumentParser(description="CTGAN Command Line Interface")
    parser.add_argument("-e", "--epochs", default=300, type=int, help="Number of training epochs")
    parser.add_argument(
        "-t", "--tsv", action="store_true", help="Load data in TSV format instead of CSV"
    )
    parser.add_argument(
        "--no-header",
        dest="header",
        action="store_false",
        help="The CSV file has no header. Discrete columns will be indices.",
    )

    parser.add_argument("-m", "--metadata", help="Path to the metadata")
    parser.add_argument(
        "-d", "--discrete", help="Comma separated list of discrete columns without whitespaces."
    )
    parser.add_argument(
        "-n",
        "--num-samples",
        type=int,
        help="Number of rows to sample. Defaults to the training data size",
    )

    parser.add_argument(
        "--generator_lr", type=float, default=2e-4, help="Learning rate for the generator."
    )
    parser.add_argument(
        "--discriminator_lr", type=float, default=2e-4, help="Learning rate for the discriminator."
    )

    parser.add_argument(
        "--generator_decay", type=float, default=1e-6, help="Weight decay for the generator."
    )
    parser.add_argument(
        "--discriminator_decay", type=float, default=0, help="Weight decay for the discriminator."
    )

    parser.add_argument(
        "--embedding_dim", type=int, default=128, help="Dimension of input z to the generator."
    )
    parser.add_argument(
        "--generator_dim",
        type=str,
        default="256,256",
        help="Dimension of each generator layer. " "Comma separated integers with no whitespaces.",
    )
    parser.add_argument(
        "--discriminator_dim",
        type=str,
        default="256,256",
        help="Dimension of each discriminator layer. "
        "Comma separated integers with no whitespaces.",
    )

    parser.add_argument(
        "--batch_size", type=int, default=500, help="Batch size. Must be an even number."
    )
    parser.add_argument(
        "--save", default=None, type=str, help="A filename to save the trained synthesizer."
    )
    parser.add_argument(
        "--load", default=None, type=str, help="A filename to load a trained synthesizer."
    )

    parser.add_argument(
        "--sample_condition_column", default=None, type=str, help="Select a discrete column name."
    )
    parser.add_argument(
        "--sample_condition_column_value",
        default=None,
        type=str,
        help="Specify the value of the selected discrete column.",
    )

    parser.add_argument("data", help="Path to training data")
    parser.add_argument("output", help="Path of the output file")

    return parser.parse_args()


def main():
    """CLI."""
    args = _parse_args()
    if args.tsv:
        data, discrete_columns = read_tsv(args.data, args.metadata)
    else:
        data, discrete_columns = read_csv(args.data, args.metadata, args.header, args.discrete)

    if args.load:
        model = CTGAN.load(args.load)
    else:
        generator_dim = [int(x) for x in args.generator_dim.split(",")]
        discriminator_dim = [int(x) for x in args.discriminator_dim.split(",")]
        model = CTGAN(
            embedding_dim=args.embedding_dim,
            generator_dim=generator_dim,
            discriminator_dim=discriminator_dim,
            generator_lr=args.generator_lr,
            generator_decay=args.generator_decay,
            discriminator_lr=args.discriminator_lr,
            discriminator_decay=args.discriminator_decay,
            batch_size=args.batch_size,
            epochs=args.epochs,
        )
    model.fit(data, discrete_columns)

    if args.save is not None:
        model.save(args.save)

    num_samples = args.num_samples or len(data)

    if args.sample_condition_column is not None:
        assert args.sample_condition_column_value is not None

    sampled = model.sample(
        num_samples, args.sample_condition_column, args.sample_condition_column_value
    )

    if args.tsv:
        write_tsv(sampled, args.metadata, args.output)
    else:
        sampled.to_csv(args.output, index=False)


if __name__ == "__main__":
    main()



import pandas as pd

from sdgx.utils import logger


class PairMetric(object):
    """PairMetric
    Metrics used to evaluate the quality of synthetic data columns.
    """

    upper_bound = None
    lower_bound = None
    metric_name = "Correlation"

    def __init__(self) -> None:
        pass

    @classmethod
    def check_input(cls, src_col: pd.Series, tar_col: pd.Series, metadata: dict):
        """Input check for table input.
        Args:
            src_data(pd.Series ): the source data column.
            tar_data(pd.Series): the target data column .
            metadata(dict): The metadata that describes the data type of each column
        """
        # Input parameter must not contain None value
        if real_data is None or synthetic_data is None:
            raise TypeError("Input contains None.")
        # check column_names
        tar_name = tar_col.name
        src_name = src_col.name

        # check column_types
        if metadata[tar_name] != metadata[src_name]:
            raise TypeError("Type of Pair is Conflicting.")

        # if type is pd.Series, return directly
        if isinstance(real_data, pd.Series):
            return src_col, tar_col

        # if type is not pd.Series or pd.DataFrame tranfer it to Series
        try:
            src_col = pd.Series(src_col)
            tar_col = pd.Series(tar_col)
            return src_col, tar_col
        except Exception as e:
            logger.error(f"An error occurred while converting to pd.Series: {e}")

        return None, None

    @classmethod
    def calculate(cls, src_col: pd.Series, tar_col: pd.Series, metadata):
        """Calculate the metric value between pair-columns between real table and synthetic table.

        Args:
            src_data(pd.Series ): the source data column.
            tar_data(pd.Series): the target data column .
            metadata(dict): The metadata that describes the data type of each column
        """
        # This method should first check the input
        # such as:
        real_data, synthetic_data = PairMetric.check_input(src_col, tar_col)

        raise NotImplementedError()

    @classmethod
    def check_output(cls, raw_metric_value: float):
        """Check the output value.

        Args:
            raw_metric_value (float):  the calculated raw value of the Mutual Information.
        """
        raise NotImplementedError()

    pass



from sdgx.data_processors.base import DataProcessor


class Generator(DataProcessor):
    """
    Base class for generators.

    Generator is used to generate data based on relationships, rules and restrictions defined by metadata.

    For example, generate random numbers between 0 and 1 as a fraction.
    """



from __future__ import annotations

import functools
import socket
import threading
import time
import urllib.request
import warnings
from contextlib import closing
from pathlib import Path
from typing import Callable

import pandas as pd

from sdgx.log import logger

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

__all__ = [
    "download_demo_data",
    "get_demo_single_table",
    "cache",
    "Singleton",
    "find_free_port",
    "download_multi_table_demo_data",
    "get_demo_single_table",
    "time2int",
]
MULTI_TABLE_DEMO_DATA = {
    "rossman": {
        "parent_table": "store",
        "child_table": "train",
        "parent_url": "https://raw.githubusercontent.com/juniorcl/rossman-store-sales/main/databases/store.csv",
        "child_url": "https://raw.githubusercontent.com/juniorcl/rossman-store-sales/main/databases/train.csv",
        "parent_primary_keys": ["Store"],
        "child_primary_keys": ["Store", "Date"],
        "foreign_keys": ["Store"],
    }
}


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def download_demo_data(data_dir: str | Path = "./dataset") -> Path:
    """
    Download demo data if not exist

    Args:
        data_dir(str | Path): data directory

    Returns:
        pathlib.Path: demo data path
    """
    data_dir = Path(data_dir).expanduser().resolve()
    demo_data_path = data_dir / "adult.csv"
    if not demo_data_path.exists():
        # Download from datahub
        demo_data_path.parent.mkdir(parents=True, exist_ok=True)

        logger.info("Downloading demo data from github data source to {}".format(demo_data_path))
        url = (
            "https://raw.githubusercontent.com/saravrajavelu/Adult-Income-Analysis/master/adult.csv"
        )
        urllib.request.urlretrieve(url, demo_data_path)

    return demo_data_path


def get_demo_single_table(data_dir: str | Path = "./dataset"):
    """
    Get demo single table as DataFrame and discrete columns names

    Args:
        data_dir(str | Path): data directory

    Returns:

        pd.DataFrame: demo single table
        list: discrete columns
    """
    demo_data_path = download_demo_data(data_dir)
    pd_obj = pd.read_csv(demo_data_path)
    discrete_cols = [
        "workclass",
        "education",
        "marital-status",
        "occupation",
        "relationship",
        "race",
        "gender",
        "native-country",
        "income",
    ]
    return pd_obj, discrete_cols


def time2int(datetime, form="%Y-%m-%d %H:%M:%S"):
    time_array = time.strptime(str(datetime), form)
    time_stamp = int(time.mktime(time_array))
    return time_stamp


class Singleton(type):
    """
    metaclass for singleton, thread-safe.
    """

    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def download_multi_table_demo_data(
    data_dir: str | Path = "./dataset", dataset_name="rossman"
) -> dict[str, Path]:
    """
    Download multi-table demo data "Rossman Store Sales" or "Rossmann Store Sales" if not exist

    Args:
        data_dir(str | Path): data directory

    Returns:
        dict[str, pathlib.Path]: dict, the key is table name, value is demo data path
    """
    demo_data_info = MULTI_TABLE_DEMO_DATA[dataset_name]
    data_dir = Path(data_dir).expanduser().resolve()
    parent_file_name = dataset_name + "_" + demo_data_info["parent_table"] + ".csv"
    child_file_name = dataset_name + "_" + demo_data_info["child_table"] + ".csv"
    demo_data_path_parent = data_dir / parent_file_name
    demo_data_path_child = data_dir / child_file_name
    # For now, I think it's OK to hardcode the URL for each dataset
    # In the future we can consider using our own S3 Bucket or providing more data sets through sdg.idslab.io.
    if not demo_data_path_parent.exists():
        # make dir
        demo_data_path_parent.parent.mkdir(parents=True, exist_ok=True)
        # download parent table from github link
        logger.info("Downloading parent table from github to {}".format(demo_data_path_parent))
        parent_url = demo_data_info["parent_url"]
        urllib.request.urlretrieve(parent_url, demo_data_path_parent)
    # then child table
    if not demo_data_path_child.exists():
        # make dir
        demo_data_path_child.parent.mkdir(parents=True, exist_ok=True)
        # download child table from github link
        logger.info("Downloading child table from github to {}".format(demo_data_path_child))
        parent_url = demo_data_info["child_url"]
        urllib.request.urlretrieve(parent_url, demo_data_path_child)

    return {
        demo_data_info["parent_table"]: demo_data_path_parent,
        demo_data_info["child_table"]: demo_data_path_child,
    }


def get_demo_multi_table(
    data_dir: str | Path = "./dataset", dataset_name="rossman"
) -> dict[str, pd.DataFrame]:
    """
    Get multi-table demo data as DataFrame and relationship

    Args:
        data_dir(str | Path): data directory

    Returns:
        dict[str, pd.DataFrame]: multi-table data dict, the key is table name, value is DataFrame.
    """
    multi_table_dict = {}
    # download if not exist
    demo_data_dict = download_multi_table_demo_data(data_dir, dataset_name)
    # read Data from path
    for table_name in demo_data_dict.keys():
        each_path = demo_data_dict[table_name]
        pd_obj = pd.read_csv(each_path)
        multi_table_dict[table_name] = pd_obj

    return multi_table_dict


def ignore_warnings(category: Warning):
    def ignore_warnings_decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=category)
                return func(*args, **kwargs)

        return wrapper

    return ignore_warnings_decorator



from __future__ import annotations

from typing import Any

import pandas as pd
from faker import Faker

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.generators.pii import PIIGenerator

fake = Faker(locale="zh_CN")


class ChnPiiGenerator(PIIGenerator):
    """ """

    chn_id_columns_list: list

    chn_phone_columns_list: list

    chn_name_columns_list: list

    chn_company_name_list: list

    fitted: bool

    def __init__(self):
        self.chn_id_columns_list = []
        self.chn_phone_columns_list = []
        self.chn_name_columns_list = []
        self.chn_company_name_list = []
        self.fitted = False

    @property
    def chn_pii_columns(self):
        return (
            self.chn_id_columns_list
            + self.chn_name_columns_list
            + self.chn_phone_columns_list
            + self.chn_company_name_list
        )

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):

        for each_col in metadata.column_list:
            data_type = metadata.get_column_data_type(each_col)
            if data_type == "chinese_name":
                self.chn_name_columns_list.append(each_col)
                continue
            if data_type == "china_mainland_mobile_phone":
                self.chn_phone_columns_list.append(each_col)
                continue
            if data_type == "china_mainland_id":
                self.chn_id_columns_list.append(each_col)
                continue
            if data_type == "chinese_company_name":
                self.chn_company_name_list.append(each_col)

        self.fitted = True

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        # if empty, return directly
        if not self.chn_pii_columns:
            return raw_data
        processed_data = raw_data
        # remove every chn pii column from the dataframe
        for each_col in self.chn_pii_columns:
            processed_data = self.remove_columns(processed_data, each_col)
        return processed_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        # if empty, return directly
        if not self.chn_pii_columns:
            return processed_data

        df_length = processed_data.shape[0]

        # chn id
        for each_col_name in self.chn_id_columns_list:
            each_email_col = [fake.ssn() for _ in range(df_length)]
            each_email_df = pd.DataFrame({each_col_name: each_email_col})
            processed_data = self.attach_columns(processed_data, each_email_df)
        # chn phone
        for each_col_name in self.chn_phone_columns_list:
            each_email_col = [fake.phone_number() for _ in range(df_length)]
            each_email_df = pd.DataFrame({each_col_name: each_email_col})
            processed_data = self.attach_columns(processed_data, each_email_df)
        # chn name
        for each_col_name in self.chn_name_columns_list:
            each_email_col = [fake.name() for _ in range(df_length)]
            each_email_df = pd.DataFrame({each_col_name: each_email_col})
            processed_data = self.attach_columns(processed_data, each_email_df)
        # chn company
        for each_col_name in self.chn_company_name_list:
            each_company_col = [fake.company() for _ in range(df_length)]
            each_company_df = pd.DataFrame({each_col_name: each_company_col})
            processed_data = self.attach_columns(processed_data, each_company_df)

        return processed_data


@hookimpl
def register(manager):
    manager.register("chnpiigenerator", ChnPiiGenerator)



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Generator

import pandas as pd

from sdgx.data_exporters.base import DataExporter
from sdgx.exceptions import CannotExportError


class CsvExporter(DataExporter):
    def __init__(self, **kwargs):
        self.to_csv_kwargs = kwargs
        if "header" in self.to_csv_kwargs:
            self.to_csv_kwargs.pop("header")
        if "index" in self.to_csv_kwargs:
            self.to_csv_kwargs.pop("index")

    def write(
        self,
        dst: str | Path,
        data: pd.DataFrame | Generator[pd.DataFrame, None, None],
    ) -> None:
        if isinstance(data, pd.DataFrame):
            data.to_csv(dst, index=False, **self.to_csv_kwargs)
        elif isinstance(data, Generator):
            with open(dst, "a") as file:
                for df in data:
                    df.to_csv(file, header=file.tell() == 0, index=False, **self.to_csv_kwargs)
        else:
            raise CannotExportError(f"Cannot export data of type {type(data)} to csv")


from sdgx.data_exporters.extension import hookimpl


@hookimpl
def register(manager):
    manager.register("CsvExporter", CsvExporter)



"""Transformer for boolean data."""

import numpy as np
import pandas as pd

from sdgx.models.components.sdv_rdt.transformers.base import BaseTransformer
from sdgx.models.components.sdv_rdt.transformers.null import NullTransformer


class BinaryEncoder(BaseTransformer):
    """Transformer for boolean data.

    This transformer replaces boolean values with their integer representation
    transformed to float.

    Null values are replaced using a ``NullTransformer``.

    Args:
        missing_value_replacement (object or None):
            Indicate what to do with the null values. If an object is given, replace them
            with the given value. If the string ``'mode'`` is given, replace them with the
            most common value. If ``None`` is given, do not replace them.
            Defaults to ``None``.
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
    """

    INPUT_SDTYPE = "boolean"
    DETERMINISTIC_TRANSFORM = True
    DETERMINISTIC_REVERSE = True

    null_transformer = None

    def __init__(self, missing_value_replacement=None, model_missing_values=False):
        self.missing_value_replacement = missing_value_replacement
        self.model_missing_values = model_missing_values

    def get_output_sdtypes(self):
        """Return the output sdtypes returned by this transformer.

        Returns:
            dict:
                Mapping from the transformed column names to the produced sdtypes.
        """
        output_sdtypes = {
            "value": "float",
        }
        if self.null_transformer and self.null_transformer.models_missing_values():
            output_sdtypes["is_null"] = "float"

        return self._add_prefix(output_sdtypes)

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit to.
        """
        self.null_transformer = NullTransformer(
            self.missing_value_replacement, self.model_missing_values
        )
        self.null_transformer.fit(data)

    def _transform(self, data):
        """Transform boolean to float.

        The boolean values will be replaced by the corresponding integer
        representations as float values.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns
            pandas.DataFrame or pandas.Series
        """
        data = pd.to_numeric(data, errors="coerce")
        return self.null_transformer.transform(data).astype(float)

    def _reverse_transform(self, data):
        """Transform float values back to the original boolean values.

        Args:
            data (pandas.DataFrame or pandas.Series):
                Data to revert.

        Returns:
            pandas.Series:
                Reverted data.
        """
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        if self.missing_value_replacement is not None:
            data = self.null_transformer.reverse_transform(data)

        if isinstance(data, np.ndarray):
            if data.ndim == 2:
                data = data[:, 0]

            data = pd.Series(data)

        isna = data.isna()
        data = np.round(data).clip(0, 1).astype("boolean").astype("object")
        data[isna] = np.nan

        return data



from __future__ import annotations

import random
from typing import Any

import pandas as pd

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.transformers.base import Transformer
from sdgx.utils import logger


class FixedCombinationTransformer(Transformer):
    """
    A transformer that handles columns with fixed combinations in a DataFrame.

    This transformer goal to auto identifies and processes columns that have fixed relationships (high covariance) in
    a given DataFrame.

    The relationships between columns include:
      - Numerical function relationships: assess them based on covariance between the columns.
      - Categorical mapping relationships: check for duplicate values for each column.

    Note that we support one-to-one mappings between columns now, and each corresponding relationship will not
    include duplicate columns.

    For example:
    we detect that,
    1 numerical relationship: (key1, Value1, Value2)
    3 one-to-one relationships: (key1, Key2) , (Category1, Category2)

    | Key1 | Key2 | Category1 | Category2 | Value1 | Value2 |
    | :--: | :--: | :-------: | :-------: | :----: | :----: |
    |  1   |  A   |   1001   |   Apple   |   10   |   20   |
    |  2   |  B   |   1002   | Broccoli  |   15   |   30   |
    |  2   |  B   |   1001   |  Apple   |   20   |   20   |
    """

    fixed_combinations: dict[str, set[str]]
    """
    A dictionary mapping column names to sets of column names that have fixed relationships with them.
    """

    simplified_fixed_combinations: dict[str, set[str]]
    """
    A dictionary mapping column names to sets of column names that have fixed relationships with them.
    """

    column_mappings: dict[(str, str), dict[str, str]]
    """
    A dictionary mapping tuples of column names to dictionaries of value mappings.
    """

    is_been_specified: bool
    """
    A boolean that flag if exist specific combinations by user.
    If true, needn't running this auto detect transform.
    """

    def __init__(self):
        super().__init__()
        self.fixed_combinations: dict[str, set[str]] = {}
        self.simplified_fixed_combinations: dict[str, set[str]] = {}
        self.column_mappings: dict[(str, str), dict[str, str]] = {}
        self.is_been_specified = False

    @property
    def is_exist_fixed_combinations(self) -> bool:
        """
        A boolean that flag if inspector have inspected some fixed combinations.
        If False, needn't running this auto detect transform.
        """
        return bool(self.fixed_combinations)

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):
        """Fit the transformer and save the relationships between columns.

        Args:
            metadata (Metadata): Metadata object
        """
        # Check if exist specific combinations by user. If True, needn't run this auto-detect transform.
        if metadata.get("specific_combinations"):
            logger.info(
                "Fit data using FixedCombinationTransformer(been specified)... Finished (No action)."
            )
            self.is_been_specified = True
            self.fitted = True
            return

        # Check if exist fixed combinations, if not exist, needn't run this auto-detect transform.
        self.fixed_combinations = metadata.get("fixed_combinations") or dict()
        if not self.is_exist_fixed_combinations:
            logger.info(
                "Fit data using FixedCombinationTransformer(not existed)... Finished (No action)."
            )
            self.fitted = True
            return

        # Simplify the fixed_combinations, remove the symmetric and duplicate combinations
        simplified_fixed_combinations = {}
        seen = set()

        if not isinstance(self.fixed_combinations, dict):
            raise TypeError(
                "fixed_combinations should be a dict, rather than {}".format(
                    type(self.fixed_combinations).__name__
                )
            )

        for base_col, related_cols in self.fixed_combinations.items():
            # create an immutable set of base_col and related_cols
            combination = frozenset([base_col]) | frozenset(related_cols)

            # if the combination has not been seen, add it to the simplified_fixed_combinations
            if combination not in seen:
                simplified_fixed_combinations[base_col] = related_cols
                seen.add(combination)

        self.simplified_fixed_combinations = simplified_fixed_combinations
        self.has_column_mappings = False
        self.fitted = True

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """Convert the input DataFrame by identifying and storing fixed column relationships.

        This method analyzes the relationships between columns specified in simplified_fixed_combinations
        and stores their value mappings. The mappings are only computed once for the first batch of data
        to optimize performance.

        NOTE:
            TODO-Enhance-Refactor Inspector by chain-of-responsibility, base one-to-one on Identified discrete_columns.
            The current implementation has space for optimization:
            - The column_mappings definition depends on the first batch of data from the DataLoader
            - This might miss some edge cases where column relationships are very comprehensive
              (e.g., some column correspondences might only appear in later batches)
            - While processing each batch separately could avoid this issue, it would incur
              significant performance overhead
            - The current function is sufficient for most scenarios
            - In the future, we may introduce parameters to control these strategies

        Args:
            raw_data (pd.DataFrame): The input DataFrame to be processed

        Returns:
            pd.DataFrame: The processed DataFrame (unchanged in this implementation)
        """

        if self.is_been_specified:
            logger.info(
                "Converting data using FixedCombinationTransformer(been specified)... Finished (No action)."
            )
            return raw_data

        if not self.is_exist_fixed_combinations:
            logger.info(
                "Converting data using FixedCombinationTransformer(not existed)... Finished (No action)."
            )
            return raw_data

        if self.has_column_mappings:
            logger.info(
                "Converting data using FixedCombinationTransformer... Finished (No action)."
            )
            return raw_data

        logger.info("Converting data using FixedCombinationTransformer... ")

        # iterate through all simplified fixed combination relationships
        for base_col, related_cols in self.simplified_fixed_combinations.items():
            if base_col not in raw_data.columns:
                continue

            # get the unique values of the base column
            base_values = raw_data[base_col].unique()

            # process each related column
            for related_col in related_cols:
                if related_col not in raw_data.columns:
                    continue

                # creating a value mapping dictionary
                value_mapping = {}
                for base_val in base_values:
                    # create a value mapping dictionary
                    related_vals = raw_data[raw_data[base_col] == base_val][related_col].unique()
                    if len(related_vals) == 1:
                        value_mapping[base_val] = related_vals[0]

                # Save when the mapping dictionary is non-empty and the reference column is not entirely NaN.
                if value_mapping and not any(pd.isna(v) for v in value_mapping.values()):
                    self.column_mappings[(base_col, related_col)] = value_mapping
                    logger.debug(f"Saved mapping relationship between {base_col} and {related_col}")

        logger.info("Converting data using FixedCombinationTransformer... Finished.")

        self.has_column_mappings = True

        return raw_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        Reverses the conversion process applied by the FixedCombinationTransformer.

        This method takes the processed DataFrame and uses the saved column mappings
        to restore the original values based on the relationships defined during the
        conversion process. If a base value does not have a corresponding related value,
        a random base value is selected to ensure that the DataFrame remains consistent.

        Args:
            processed_data (pd.DataFrame): The input DataFrame containing the processed data.

        Returns:
            pd.DataFrame: The DataFrame with original values restored based on the defined mappings.
        """
        if self.is_been_specified:
            logger.info(
                "Reverse converting data using FixedCombinationTransformer(been specified)... Finished (No action)."
            )
            return processed_data

        if not self.is_exist_fixed_combinations:
            logger.info(
                "Reverse converting data using FixedCombinationTransformer(not existed)... Finished (No action)."
            )
            return processed_data

        result_df = processed_data.copy()

        logger.info("Reverse converting data using FixedCombinationTransformer...")

        # iterate through all column mappings
        for (base_col, related_col), mapping in self.column_mappings.items():
            if base_col not in result_df.columns or related_col not in result_df.columns:
                continue

            # define a function to replace base_col and related_col
            def replace_row(row):
                base_val = row[base_col]
                if base_val in mapping:
                    new_related_val = mapping[base_val]
                    return pd.Series({base_col: base_val, related_col: new_related_val})
                else:
                    # randomly select a base_val and get the corresponding related_val
                    new_base_val = random.choice(list(mapping.keys()))
                    new_related_val = mapping[new_base_val]
                    return pd.Series({base_col: new_base_val, related_col: new_related_val})

            # apply the `replace` function, and update the DataFrame
            replaced = result_df.apply(replace_row, axis=1)
            result_df[base_col] = replaced[base_col]
            result_df[related_col] = replaced[related_col]

        logger.info("Reverse converting data using FixedCombinationTransformer... Finished.")
        return result_df


@hookimpl
def register(manager):
    manager.register("FixedCombinationTransformer", FixedCombinationTransformer)



from __future__ import annotations

import time
from pathlib import Path
from typing import Any, Generator

import pandas as pd
from tqdm import autonotebook as tqdm

from sdgx.data_connectors.base import DataConnector
from sdgx.data_connectors.generator_connector import GeneratorConnector
from sdgx.data_connectors.manager import DataConnectorManager
from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.base import DataProcessor
from sdgx.data_processors.manager import DataProcessorManager
from sdgx.exceptions import SynthesizerInitError, SynthesizerSampleError
from sdgx.models.base import SynthesizerModel
from sdgx.models.components.sdv_ctgan.synthesizers.base import BatchedSynthesizer
from sdgx.models.manager import ModelManager
from sdgx.models.ml.single_table.ctgan import CTGANSynthesizerModel
from sdgx.models.statistics.single_table.base import StatisticSynthesizerModel
from sdgx.utils import logger


class Synthesizer:
    """
    Synthesizer is the high level interface for synthesizing data.

    We provided several example usage in our `Github repository <https://github.com/hitsz-ids/synthetic-data-generator/tree/main/example>`_.

    Args:

        model (str | SynthesizerModel | type[SynthesizerModel]): The name of the model or the model itself. Type of model must be :class:`~sdgx.models.base.SynthesizerModel`.
            When model is a string, it must be registered in :class:`~sdgx.models.manager.ModelManager`.
        model_path (str | Path, optional): The path to the model file. Defaults to None. Used to load the model if ``model`` is a string or type of :class:`~sdgx.models.base.SynthesizerModel`.
        model_kwargs (dict[str, Any], optional): The keyword arguments for model. Defaults to None.
        metadata (Metadata, optional): The metadata to use. Defaults to None.
        metadata_path (str | Path, optional): The path to the metadata file. Defaults to None. Used to load the metadata if ``metadata`` is None.
        data_connector (DataConnector | type[DataConnector] | str, optional): The data connector to use. Defaults to None.
            When data_connector is a string, it must be registered in :class:`~sdgx.data_connectors.manager.DataConnectorManager`.
        data_connector_kwargs (dict[str, Any], optional): The keyword arguments for data connectors. Defaults to None.
        raw_data_loaders_kwargs (dict[str, Any], optional): The keyword arguments for raw data loaders. Defaults to None.
        processed_data_loaders_kwargs (dict[str, Any], optional): The keyword arguments for processed data loaders. Defaults to None.
        data_processors (list[str | DataProcessor | type[DataProcessor]], optional): The data processors to use. Defaults to None.
            When data_processor is a string, it must be registered in :class:`~sdgx.data_processors.manager.DataProcessorManager`.
        data_processors_kwargs (dict[str, dict[str, Any]], optional): The keyword arguments for data processors. Defaults to None.

    Example:

        .. code-block:: python

            from sdgx.data_connectors.csv_connector import CsvConnector
            from sdgx.models.ml.single_table.ctgan import CTGANSynthesizerModel
            from sdgx.synthesizer import Synthesizer
            from sdgx.utils import download_demo_data

            dataset_csv = download_demo_data()
            data_connector = CsvConnector(path=dataset_csv)
            synthesizer = Synthesizer(
                model=CTGANSynthesizerModel(epochs=1),  # For quick demo
                data_connector=data_connector,
            )
            synthesizer.fit()
            sampled_data = synthesizer.sample(1000)
    """

    METADATA_SAVE_NAME = "metadata.json"
    """
    Default name for metadata file
    """
    MODEL_SAVE_DIR = "model"
    """
    Default name for model directory
    """

    def __init__(
        self,
        model: str | SynthesizerModel | type[SynthesizerModel],
        model_path: None | str | Path = None,
        model_kwargs: None | dict[str, Any] = None,
        metadata: None | Metadata = None,
        metadata_path: None | str | Path = None,
        data_connector: None | str | DataConnector | type[DataConnector] = None,
        data_connector_kwargs: None | dict[str, Any] = None,
        raw_data_loaders_kwargs: None | dict[str, Any] = None,
        processed_data_loaders_kwargs: None | dict[str, Any] = None,
        data_processors: None | list[str | DataProcessor | type[DataProcessor]] = None,
        data_processors_kwargs: None | dict[str, Any] = None,
    ):
        # Init data connectors
        if isinstance(data_connector, str) or isinstance(data_connector, type):
            data_connector = DataConnectorManager().init_data_connector(
                data_connector, **(data_connector_kwargs or {})
            )
        if data_connector:
            self.dataloader = DataLoader(
                data_connector,
                **(raw_data_loaders_kwargs or {}),
            )
        else:
            logger.warning("No data_connector provided, will not support `fit`")
            self.dataloader = None

        # Init data processors
        self.data_processors_manager = DataProcessorManager()
        if not data_processors:
            data_processors = self.data_processors_manager.registed_default_processor_list
        logger.info(f"Using data processors: {data_processors}")
        self.data_processors = [
            (
                d
                if isinstance(d, DataProcessor)
                else self.data_processors_manager.init_data_processor(
                    d, **(data_processors_kwargs or {})
                )
            )
            for d in data_processors
        ]
        if metadata and metadata_path:
            raise SynthesizerInitError(
                "metadata and metadata_path cannot be specified at the same time"
            )

        # Load metadata
        # metadata also can be changed in ``fit`` or ``sample``
        # Always use the latest metadata configured.
        if metadata:
            self.metadata = metadata
        elif metadata_path:
            self.metadata = Metadata.load(metadata_path)
        else:
            self.metadata = None

        # Init model
        self.model_manager = ModelManager()
        if isinstance(model, SynthesizerModel) and model_path:
            # Initialized model cannot load from model_path
            raise SynthesizerInitError(
                "model as instance and model_path cannot be specified at the same time"
            )
        if (isinstance(model, str) or isinstance(model, type)) and model_path:
            # Load model by cls or str
            self.model = self.model_manager.load(model, model_path, **(model_kwargs or {}))
            if model_kwargs:
                logger.warning("model_kwargs will be ignored when loading model from model_path")
        elif isinstance(model, str) or isinstance(model, type):
            # Init model by cls or str
            self.model = self.model_manager.init_model(model, **(model_kwargs or {}))
        elif isinstance(model, SynthesizerModel) or isinstance(model, StatisticSynthesizerModel):
            # Already initialized model
            self.model = model
            if model_kwargs:
                logger.warning("model_kwargs will be ignored when using already initialized model")
        else:
            raise SynthesizerInitError("model or model_path must be specified")

        # Other arguments
        self.processed_data_loaders_kwargs = processed_data_loaders_kwargs or {}

    def save(self, save_dir: str | Path) -> Path:
        """
        Dump metadata and model to file

        Args:
            save_dir (str | Path): The directory to save the model.

        Returns:
            Path: The directory to save the synthesizer.
        """
        save_dir = Path(save_dir).expanduser().resolve()
        save_dir.mkdir(parents=True, exist_ok=True)
        logger.info(f"Saving synthesizer to {save_dir}")

        if self.metadata:
            self.metadata.save(save_dir / self.METADATA_SAVE_NAME)

        model_save_dir = save_dir / self.MODEL_SAVE_DIR
        model_save_dir.mkdir(parents=True, exist_ok=True)
        self.model.save(model_save_dir)
        return save_dir

    @classmethod
    def load(
        cls,
        load_dir: str | Path,
        model: str | type[SynthesizerModel],
        metadata: None | Metadata = None,
        data_connector: None | str | DataConnector | type[DataConnector] = None,
        data_connector_kwargs: None | dict[str, Any] = None,
        raw_data_loaders_kwargs: None | dict[str, Any] = None,
        processed_data_loaders_kwargs: None | dict[str, Any] = None,
        data_processors: None | list[str | DataProcessor | type[DataProcessor]] = None,
        data_processors_kwargs: None | dict[str, dict[str, Any]] = None,
        model_kwargs=None,
    ) -> "Synthesizer":
        """
        Load metadata and model, allow rebuilding Synthesizer for finetuning or other use cases.

        We need ``model`` as not every model support *pickle* way to save and load.

        Args:
            load_dir (str | Path): The directory to load the model.
            model (str | type[SynthesizerModel]): The name of the model or the model itself. Type of model must be :class:`~sdgx.models.base.SynthesizerModel`.
                When model is a string, it must be registered in :class:`~sdgx.models.manager.ModelManager`.
            metadata (Metadata, optional): The metadata to use. Defaults to None.
            data_connector (DataConnector | type[DataConnector] | str, optional): The data connector to use. Defaults to None.
                When data_connector is a string, it must be registered in :class:`~sdgx.data_connectors.manager.DataConnectorManager`.
            data_connector_kwargs (dict[str, Any], optional): The keyword arguments for data connectors. Defaults to None.
            raw_data_loaders_kwargs (dict[str, Any], optional): The keyword arguments for raw data loaders. Defaults to None.
            processed_data_loaders_kwargs (dict[str, Any], optional): The keyword arguments for processed data loaders. Defaults to None.
            data_processors (list[str | DataProcessor | type[DataProcessor]], optional): The data processors to use. Defaults to None.
                When data_processor is a string, it must be registered in :class:`~sdgx.data_processors.manager.DataProcessorManager`.
            data_processors_kwargs (dict[str, dict[str, Any]], optional): The keyword arguments for data processors. Defaults to None.

        Returns:
            Synthesizer: The synthesizer instance.
        """

        load_dir = Path(load_dir).expanduser().resolve()
        logger.info(f"Loading synthesizer from {load_dir}")

        if not load_dir.exists():
            raise SynthesizerInitError(f"{load_dir.as_posix()} does not exist")
        model_path = load_dir / cls.MODEL_SAVE_DIR
        if not model_path.exists():
            raise SynthesizerInitError(
                f"{model_path.as_posix()} does not exist, cannot load model."
            )

        metadata_path = load_dir / cls.METADATA_SAVE_NAME
        if not metadata_path.exists():
            metadata_path = None

        return Synthesizer(
            model=model,
            model_path=model_path,
            metadata=metadata,
            metadata_path=metadata_path,
            model_kwargs=model_kwargs,
            data_connector=data_connector,
            data_connector_kwargs=data_connector_kwargs,
            raw_data_loaders_kwargs=raw_data_loaders_kwargs,
            processed_data_loaders_kwargs=processed_data_loaders_kwargs,
            data_processors=data_processors,
            data_processors_kwargs=data_processors_kwargs,
        )

    def fit(
        self,
        metadata: None | Metadata = None,
        inspector_max_chunk: int = 10,
        metadata_include_inspectors: None | list[str] = None,
        metadata_exclude_inspectors: None | list[str] = None,
        inspector_init_kwargs: None | dict[str, Any] = None,
        model_fit_kwargs: None | dict[str, Any] = None,
    ):
        """
        Fit the synthesizer with metadata and data processors.

        Raw data will be loaded from the dataloader and processed by the data processors in a Generator.
        The Generator, which prevents the processed data, will be wrapped into a DataLoader, aka ProcessedDataLoader.
        The ProcessedDataLoader will be used to fit the model.

        For more information about DataLoaders, please refer to the :class:`~sdgx.data_loaders.base.DataLoader`.

        For more information about DataProcessors, please refer to the :class:`~sdgx.data_processors.base.DataProcessor`.

        For more information about DataConnectors, please refer to the :class:`~sdgx.data_connectors.base.DataConnector`. Especially, the :class:`~sdgx.data_connectors.generator_connector.GeneratorConnector`.

        Args:
            metadata (Metadata, optional): The metadata to use. Defaults to None. If None, it will be inferred from the dataloader with the :func:`~sdgx.data_models.metadata.Metadata.from_dataloader` method.
            inspector_max_chunk (int, optional): The maximum number of chunks to inspect. Defaults to 10.
            metadata_include_inspectors (list[str], optional): The list of metadata inspectors to include. Defaults to None.
            metadata_exclude_inspectors (list[str], optional): The list of metadata inspectors to exclude. Defaults to None.
            inspector_init_kwargs (dict[str, Any], optional): The keyword arguments for metadata inspectors. Defaults to None.
            model_fit_kwargs (dict[str, Any], optional): The keyword arguments for model.fit. Defaults to None.
        """
        if self.dataloader is None:
            raise SynthesizerInitError(
                "Cannot fit without dataloader, check `data_connector` parameter when initializing Synthesizer"
            )

        metadata = (
            metadata
            or self.metadata
            or Metadata.from_dataloader(
                self.dataloader,
                max_chunk=inspector_max_chunk,
                include_inspectors=metadata_include_inspectors,
                exclude_inspectors=metadata_exclude_inspectors,
                inspector_init_kwargs=inspector_init_kwargs,
            )
        )
        # Some processors may cause metadata update before model fitting, we need to make a copy.
        self.metadata = metadata.model_copy()  # Ensure update metadata

        logger.info("Fitting data processors...")
        if not self.dataloader:
            logger.info("Fitting without dataloader.")
        start_time = time.time()
        for d in self.data_processors:
            if self.dataloader:
                d.fit(metadata=metadata, tabular_data=self.dataloader)
            else:
                d.fit(metadata=metadata)
        logger.info(
            f"Fitted {len(self.data_processors)} data processors in  {time.time() - start_time}s."
        )

        def chunk_generator() -> Generator[pd.DataFrame, None, None]:
            for chunk in self.dataloader.iter():
                for d in self.data_processors:
                    chunk = d.convert(chunk)
                yield chunk

        logger.info("Initializing processed data loader...")
        start_time = time.time()
        processed_dataloader = DataLoader(
            GeneratorConnector(chunk_generator),
            identity=self.dataloader.identity,
            **self.processed_data_loaders_kwargs,
        )
        logger.info(f"Initialized processed data loader in {time.time() - start_time}s")
        try:
            logger.info("Model fit Started...")
            self.model.fit(metadata, processed_dataloader, **(model_fit_kwargs or {}))
            logger.info("Model fit... Finished")
        finally:
            processed_dataloader.finalize(clear_cache=True)

    def sample(
        self,
        count: int,
        chunksize: None | int = None,
        metadata: None | Metadata = None,
        model_sample_args: None | dict[str, Any] = None,
    ) -> pd.DataFrame | Generator[pd.DataFrame, None, None]:
        """
        Sample data from the synthesizer.

        Args:
            count (int): The number of samples to generate.
            chunksize (int, optional): The chunksize to use. Defaults to None. If is not None, the data will be sampled in chunks.
                And will return a generator that yields chunks of samples.
            metadata (Metadata, optional): The metadata to use. Defaults to None. If None, will use the metadata in fit first.
            model_sample_args (dict[str, Any], optional): The keyword arguments for model.sample. Defaults to None.

        Returns:
            pd.DataFrame | typing.Generator[pd.DataFrame, None, None]: The sampled data. When chunksize is not None, it will be a generator.
        """
        logger.info("Sampling...")
        metadata = metadata or self.metadata
        self.metadata = metadata  # Ensure update metadata

        # data_processors do not need to be fit again in the sampling stage

        if not model_sample_args:
            model_sample_args = {}

        if chunksize is None:
            return self._sample_once(count, model_sample_args)

        if chunksize > count:
            raise SynthesizerSampleError("chunksize must be less than or equal to count")

        def generator_sample_caller():
            sample_times = count // chunksize
            for _ in range(sample_times):
                sample_data = self._sample_once(chunksize, model_sample_args)
                for d in self.data_processors:
                    sample_data = d.reverse_convert(sample_data)
                yield sample_data

            if count % chunksize > 0:
                sample_data = self._sample_once(count % chunksize, model_sample_args)
                for d in self.data_processors:
                    sample_data = d.reverse_convert(sample_data)
                yield sample_data

        return generator_sample_caller()

    def _sample_once(
        self, count: int, model_sample_args: None | dict[str, Any] = None
    ) -> pd.DataFrame:
        """
        Sample data once.

        DataProcessors may drop some broken data after reverse_convert.
        So we oversample first and then take the first `count` samples.

        TODO:

            - Use an adaptive scale for oversampling will be better for performance.

        """
        missing_count = count
        max_trails = 50
        sample_data_list = []
        psb = tqdm.tqdm(total=count, desc="Sampling")

        # To improve batched model performance, such as tvae or ctgan.
        batch_size: int = 0
        multiply_factor: float = 4.0
        if isinstance(self.model, BatchedSynthesizer):
            batch_size = self.model.get_batch_size()
            multiply_factor = 1.2
            if isinstance(self.model, CTGANSynthesizerModel):
                model_sample_args = {"drop_more": False}

        while missing_count > 0 and max_trails > 0:
            sample_data = self.model.sample(
                max(int(missing_count * multiply_factor), batch_size), **model_sample_args
            )
            # TODO table separated parallel process
            for d in self.data_processors:
                sample_data = d.reverse_convert(sample_data)
            sample_data = sample_data.dropna(how="all")
            sample_data_list.append(sample_data)
            missing_count = missing_count - len(sample_data)
            psb.update(len(sample_data))
            max_trails -= 1

        return pd.concat(sample_data_list)[:count]

    def cleanup(self):
        """
        Cleanup resources. This will cause model unavailable and clear the cache.

        It useful when Synthesizer object is no longer needed and may hold large resources like GPUs.
        """

        if self.dataloader:
            self.dataloader.finalize(clear_cache=True)
        # Release resources
        if hasattr(self, "model"):
            del self.model

    def __del__(self):
        self.cleanup()



from __future__ import annotations

from typing import Any, List

import pandas as pd

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.formatters.base import Formatter
from sdgx.utils import logger


class IntValueFormatter(Formatter):
    """
    Formatter class for handling Int values in pd.DataFrame.
    """

    int_columns: set
    """
    List of column names that are of type int, populated by the fit method using metadata.
    """

    def __init__(self):
        self.int_columns = set()

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):
        """
        Fit method for the formatter.

        Formatter need to use metadata to record which columns belong to the int type, and convert them back to the int type during post-processing.
        """

        # get from metadata
        for each_col in metadata.int_columns:
            if each_col not in metadata.column_list:
                continue
            if metadata.get_column_data_type(each_col) == "int":
                self.int_columns.add(each_col)
                continue
            if metadata.get_column_data_type(each_col) == "id":
                self.int_columns.add(each_col)

        logger.info("IntValueFormatter Fitted.")
        self.fitted = True

        return

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        No action for convert.
        """

        logger.info("Converting data using IntValueFormatter... Finished  (No Action).")

        return raw_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        reverse_convert method for the formatter.

        Do format conversion for int columns.
        """

        for col in self.int_columns:
            if col in processed_data.columns:
                processed_data[col] = processed_data[col].astype(int)
            else:
                logger.error("Column {} not found in processed_data.".format(col))
        logger.info("Data reverse-converted by IntValueFormatter.")

        return processed_data


@hookimpl
def register(manager):
    manager.register("IntValueFormatter", IntValueFormatter)



import numpy as np
import pandas as pd
from scipy.stats import entropy
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.preprocessing import LabelEncoder

from sdgx.metrics.pair_column.base import PairMetric
from sdgx.utils import time2int


class MISim(PairMetric):
    """MISim : Mutual Information Similarity

    This class is used to calculate the Mutual Information Similarity between the target columns of real data and synthetic data.

    Currently, we support discrete and continuous(need to be discretized) columns as inputs.
    """

    def __init__(instance) -> None:
        super().__init__()
        instance.lower_bound = 0
        instance.upper_bound = 1
        instance.metric_name = "mutual_information_similarity"
        instance.numerical_bins = 50

    @classmethod
    def calculate(
        cls,
        src_col: pd.Series,
        tar_col: pd.Series,
        metadata: dict,
    ) -> float:
        """
        Calculate the MI similarity for the source data colum and the target data column.
        Args:
            src_data(pd.Series ): the source data column.
            tar_data(pd.Series): the target data column .
            metadata(dict): The metadata that describes the data type of each columns
        Returns:
            MI_similarity (float): The metric value.
        """

        # 传入概率分布数组
        instance = cls()

        col_name = src_col.name
        data_type = metadata[col_name]

        if data_type == "numerical":
            x = np.array(src_col.array)
            src_col = pd.cut(
                x,
                instance.numerical_bins,
                labels=range(instance.numerical_bins),
            )
            x = np.array(tar_col.array)
            tar_col = pd.cut(
                x,
                instance.numerical_bins,
                labels=range(instance.numerical_bins),
            )
            src_col = src_col.to_numpy()
            tar_col = tar_col.to_numpy()

        elif data_type == "category":
            le = LabelEncoder()
            src_list = list(set(src_col.array))
            tar_list = list(set(tar_col.array))
            fit_list = tar_list + src_list
            le.fit(fit_list)

            src_col = le.transform(np.array(src_col.array))
            tar_col = le.transform(np.array(tar_col.array))

        elif data_type == "datetime":
            src_col = src_col.apply(time2int)
            tar_col = tar_col.apply(time2int)
            src_col = pd.cut(
                src_col, bins=instance.numerical_bins, labels=range(instance.numerical_bins)
            )
            tar_col = pd.cut(
                tar_col, bins=instance.numerical_bins, labels=range(instance.numerical_bins)
            )
            src_col = src_col.to_numpy()
            tar_col = tar_col.to_numpy()

        MI_sim = normalized_mutual_info_score(src_col, tar_col)
        return MI_sim

    @classmethod
    def check_output(cls, raw_metric_value: float):
        """Check the output value.

        Args:
            raw_metric_value (float):  the calculated raw value of the MI similarity.
        """
        pass



"""Synthesizers module."""

from sdgx.models.components.sdv_ctgan.synthesizers.ctgan import CTGAN
from sdgx.models.components.sdv_ctgan.synthesizers.tvae import TVAE

__all__ = ("CTGAN", "TVAE")


def get_all_synthesizers():
    return {name: globals()[name] for name in __all__}



from __future__ import annotations

from sdgx.data_exporters.base import DataExporter


class MyOwnExporter(DataExporter): ...


from sdgx.data_exporters.extension import hookimpl


@hookimpl
def register(manager):
    manager.register("MyOwnExporter", MyOwnExporter)



"""
Move from utils/io/mysql.py

TODO: Implement as MySQLConnector
"""

# import os

# import pymysql
# from config import sdg_log_dir
# from pymysql.err import IntegrityError, OperationalError

# # logger 需要解决
# from sdgx.utils import logger


# class mysql_db_connector:
#     def __init__(
#         self,
#         databaes_ip,
#         database_port,
#         database_username,
#         database_passwd,
#         database_name,
#         database_type="mysql",
#     ) -> None:
#         # 配置log路径
#         if not os.path.exists(sdg_log_dir):
#             os.makedirs(sdg_log_dir)
#         log_path = os.path.join(sdg_log_dir, "mysql_log.log")
#         log_path_handler = get_log_file_handler(log_path)
#         logger.addHandler(log_path_handler)

#         # 进行初始化
#         self.database_type = database_type
#         self.databaes_ip = databaes_ip
#         self.database_port = database_port
#         self.database_username = database_username
#         self.database_passwd = database_passwd
#         self.database_name = database_name

#         # 记录状态
#         self.conn = None
#         self.status = "waiting"
#         self.insert_success = False
#         self.del_success = False

#         # 建立一个 conn，
#         if self.database_type == "mysql":
#             logger.info(
#                 "Establish Mysql Connection with %s:%s" % (self.databaes_ip, self.database_port)
#             )
#             try:
#                 self.conn = pymysql.connect(
#                     host=self.databaes_ip,
#                     port=self.database_port,
#                     user=self.database_username,
#                     password=self.database_passwd,
#                     database=self.database_name,
#                 )
#                 self.status = "success"
#                 logger.info("MySQL Connection Created.")
#             except ConnectionRefusedError:
#                 self.status = "failed"
#                 logger.error("MySQL Connection Refused")
#             except OperationalError:
#                 logger.error("MySQL OperationalError")
#                 self.status = "failed"
#         else:
#             logger.error("Not support other Database yet.")
#             raise NotImplementedError("Not support other Database yet.")
#         pass

#     def query(self, sql):
#         # 查询动作，把数据给提出来，以列表形式返回
#         # assert 'select' in sql.lowercase()
#         if "select" not in sql.lower():
#             raise ValueError("NOT a Query SQL")
#         cursor = self.conn.cursor()
#         select_cnt = cursor.execute(sql)  # 得到结果行数
#         res = []
#         for _ in range(select_cnt):
#             sample = list(cursor.fetchone())
#             res.append(sample)
#         cursor.close()
#         logger.debug('Query: "%s", Success' % sql.replace("\n", " ").strip())
#         return res  # 结果一定是二维数组

#     def insert(self, sql):
#         # INSERT INTO `fed_sql`.`b_job` (`id`, `job_id`, `job_sql`, `job_ priority`, `grammar_check`, `integrity_check`, `status`, `create_time`, `update_time`, `start_time`) VALUES ('49', '1', 'selec * from A.xx;', '2', '1', '1', 'success', '1657871901', '1657871901', '1657871901');
#         if "insert" not in sql.lower() and "update" not in sql.lower():
#             raise ValueError("NOT a Insert SQL")
#         cursor = self.conn.cursor()
#         try:
#             cursor.execute(sql)
#             cursor.close()
#         except IntegrityError:
#             logger.error('Insert: "%s", Failed' % sql.replace("\n", " "))
#             self.insert_success = False
#         cursor.close()
#         self.conn.commit()
#         if "insert" in sql.lower():
#             logger.debug('Insert: "%s", Success' % sql.replace("\n", " "))
#         elif "update" in sql.lower():
#             logger.debug('Update: "%s", Success' % sql.replace("\n", " "))
#         self.insert_success = True

#     def delete(self, sql):
#         # DELETE FROM `fed_sql`.`s_party` WHERE (`id` = '3');
#         if "delete" not in sql.lower():
#             raise ValueError("NOT a Delete SQL")
#         cursor = self.conn.cursor()
#         try:
#             cursor.execute(sql)
#             cursor.close()
#         except Exception as e:
#             logger.error('Delete: "%s", Failed' % sql.replace("\n", " "))
#             logger.error(str(e))
#             self.del_success = False
#         cursor.close()
#         self.conn.commit()
#         logger.debug('Delete: "%s", Success' % sql.replace("\n", " "))
#         self.del_success = True

#     def get_tables(self):
#         sql = "show tables;"
#         cursor = self.conn.cursor()
#         table_cnt = cursor.execute(sql)
#         res = []
#         for _ in range(table_cnt):
#             table = cursor.fetchone()[0]
#             res.append(table)
#         return res

#     def create_table(self, sql):
#         if "create" not in sql.lower():
#             raise ValueError("NOT a Create SQL")
#         cursor = self.conn.cursor()
#         cursor.execute(sql)
#         self.conn.commit()
#         cursor.close()

#     def __del__(self):
#         # 关闭数据库连接
#         if self.status == "failed":
#             logger.info("NO MySQL Connection, Closed.")
#         elif self.status != "closed":
#             self.conn.close()
#             logger.info("MySQL Connection Closed.")

#     # 手工 close
#     # 一般不太用得到
#     def close(self):
#         self.conn.close()
#         self.status = "closed"
#         logger.info("MySQL Connection Closed in func <close>.")



"""Frank module."""

import sys

import numpy as np
import scipy.integrate as integrate
from scipy.optimize import least_squares

from sdgx.models.components.sdv_copulas import EPSILON
from sdgx.models.components.sdv_copulas.bivariate.base import Bivariate, CopulaTypes
from sdgx.models.components.sdv_copulas.bivariate.utils import split_matrix

MIN_FLOAT_LOG = np.log(sys.float_info.min)
MAX_FLOAT_LOG = np.log(sys.float_info.max)


class Frank(Bivariate):
    """Class for Frank copula model."""

    copula_type = CopulaTypes.FRANK
    theta_interval = [-float("inf"), float("inf")]
    invalid_thetas = [0]

    def generator(self, t):
        """Return the generator function."""
        a = (np.exp(-self.theta * t) - 1) / (np.exp(-self.theta) - 1)
        return -np.log(a)

    def _g(self, z):
        r"""Assist in solving the Frank copula.

        This functions encapsulates :math:`g(z) = e^{-\theta z} - 1` used on Frank copulas.

        Argument:
            z: np.ndarray

        Returns:
            np.ndarray

        """
        return np.exp(-self.theta * z) - 1

    def probability_density(self, X):
        r"""Compute probability density function for given copula family.

        The probability density(PDF) for the Frank family of copulas correspond to the formula:

        .. math:: c(U,V) = \frac{\partial^2 C(u,v)}{\partial v \partial u} =
             \frac{-\theta g(1)(1 + g(u + v))}{(g(u) g(v) + g(1)) ^ 2}

        Where the g function is defined by:

        .. math:: g(x) = e^{-\theta x} - 1

        Args:
            X: `np.ndarray`

        Returns:
            np.array: probability density

        """
        self.check_fit()

        U, V = split_matrix(X)

        if self.theta == 0:
            return U * V

        else:
            num = (-self.theta * self._g(1)) * (1 + self._g(U + V))
            aux = self._g(U) * self._g(V) + self._g(1)
            den = np.power(aux, 2)
            return num / den

    def cumulative_distribution(self, X):
        r"""Compute the cumulative distribution function for the Frank copula.

        The cumulative density(cdf), or distribution function for the Frank family of copulas
        correspond to the formula:

        .. math:: C(u,v) =  −\frac{\ln({\frac{1 + g(u) g(v)}{g(1)}})}{\theta}


        Args:
            X: `np.ndarray`

        Returns:
            np.array: cumulative distribution

        """
        self.check_fit()

        U, V = split_matrix(X)

        num = (np.exp(-self.theta * U) - 1) * (np.exp(-self.theta * V) - 1)
        den = np.exp(-self.theta) - 1

        return -1.0 / self.theta * np.log(1 + num / den)

    def percent_point(self, y, V):
        """Compute the inverse of conditional cumulative distribution :math:`C(u|v)^{-1}`.

        Args:
            y: `np.ndarray` value of :math:`C(u|v)`.
            v: `np.ndarray` given value of v.
        """
        self.check_fit()

        if self.theta == 0:
            return V

        else:
            return super().percent_point(y, V)

    def partial_derivative(self, X):
        r"""Compute partial derivative of cumulative distribution.

        The partial derivative of the copula(CDF) is the conditional CDF.

        .. math:: F(v|u) = \frac{\partial}{\partial u}C(u,v) =
            \frac{g(u)g(v) + g(v)}{g(u)g(v) + g(1)}

        Args:
            X (np.ndarray)
            y (float)

        Returns:
            np.ndarray

        """
        self.check_fit()

        U, V = split_matrix(X)

        if self.theta == 0:
            return V

        else:
            num = self._g(U) * self._g(V) + self._g(U)
            den = self._g(U) * self._g(V) + self._g(1)
            return num / den

    def compute_theta(self):
        r"""Compute theta parameter using Kendall's tau.

        On Frank copula, the relationship between tau and theta is defined by:

        .. math:: \tau = 1 − \frac{4}{\theta} + \frac{4}{\theta^2}\int_0^\theta \!
            \frac{t}{e^t -1} \mathrm{d}t.

        In order to solve it, we can simplify it as

        .. math:: 0 = 1 + \frac{4}{\theta}(D_1(\theta) - 1) - \tau

        where the function D is the Debye function of first order, defined as:

        .. math:: D_1(x) = \frac{1}{x}\int_0^x\frac{t}{e^t -1} \mathrm{d}t.

        """
        result = least_squares(self._tau_to_theta, 1, bounds=(MIN_FLOAT_LOG, MAX_FLOAT_LOG))
        return result.x[0]

    def _tau_to_theta(self, alpha):
        """Relationship between tau and theta as a solvable equation."""

        def debye(t):
            return t / (np.exp(t) - 1)

        debye_value = integrate.quad(debye, EPSILON, alpha)[0] / alpha
        return 4 * (debye_value - 1) / alpha + 1 - self.tau



# -*- coding: utf-8 -*-

"""Top-level package for ctgan."""

__author__ = "MIT Data To AI Lab"
__email__ = "dailabmit@gmail.com"
__version__ = "0.6.0"

from sdgx.models.components.sdv_ctgan.demo import load_demo
from sdgx.models.components.sdv_ctgan.synthesizers.ctgan import CTGAN
from sdgx.models.components.sdv_ctgan.synthesizers.tvae import TVAE

__all__ = ("CTGAN", "TVAE", "load_demo")



from __future__ import annotations

from typing import Any, Dict

import pandas as pd

from sdgx.data_models.metadata import Metadata
from sdgx.exceptions import SynthesizerProcessorError
from sdgx.utils import logger


class DataProcessor:
    """
    Base class for data processors.
    """

    fitted = False

    def check_fitted(self):
        """Check if the processor is fitted.

        Raises:
            SynthesizerProcessorError: If the processor is not fitted.
        """
        if not self.fitted:
            raise SynthesizerProcessorError("Processor NOT fitted.")

    def fit(self, metadata: Metadata | None = None, **kwargs: Dict[str, Any]):
        self._fit(metadata, **kwargs)
        self.fitted = True

    def _fit(self, metadata: Metadata | None = None, **kwargs: Dict[str, Any]):
        """Fit the data processor.

        Called before ``convert`` and ``reverse_convert``.

        Args:
            metadata (Metadata, optional): Metadata. Defaults to None.
        """
        return

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """Convert raw data into processed data.

        Args:
            raw_data (pd.DataFrame): Raw data

        Returns:
            pd.DataFrame: Processed data
        """
        return raw_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """Convert processed data into raw data.

        Args:
            processed_data (pd.DataFrame): Processed data

        Returns:
            pd.DataFrame: Raw data
        """
        return processed_data

    @staticmethod
    def remove_columns(tabular_data: pd.DataFrame, column_name_to_remove: list) -> pd.DataFrame:
        """
        Remove specified columns from the input tabular data.

        Args:
            - tabular_data (pd.DataFrame): Processed tabular data
            - column_name_to_remove (list): List of column names to be removed

        Returns:
            - result_data (pd.DataFrame): Tabular data with specified columns removed
        """

        # Make a copy of the input data to avoid modifying the original data
        result_data = tabular_data.copy()

        # Remove specified columns
        try:
            result_data = result_data.drop(columns=column_name_to_remove)
        except KeyError:
            logger.warning(
                "Duplicate column removal occurred, which might lead to unintended consequences."
            )

        return result_data

    @staticmethod
    def attach_columns(tabular_data: pd.DataFrame, new_columns: pd.DataFrame) -> pd.DataFrame:
        """
        Attach additional columns to an existing DataFrame.

        Args:
            - tabular_data (pd.DataFrame): The original DataFrame.
            - new_columns (pd.DataFrame): The DataFrame containing additional columns to be attached.

        Returns:
            - result_data (pd.DataFrame): The DataFrame with new_columns attached.

        Raises:
            - ValueError: If the number of rows in tabular_data and new_columns are not the same.
        """

        # Check if the number of rows in tabular_data and new_columns are the same
        if tabular_data.shape[0] != new_columns.shape[0]:
            raise ValueError("Number of rows in tabular_data and new_columns must be the same.")

        # Concatenate tabular_data and new_columns along axis 1 (columns)
        result_data = pd.concat([tabular_data, new_columns], axis=1)

        return result_data



"""Transformers for numerical data."""

import copy
import sys
import warnings

import numpy as np
import pandas as pd
import scipy
from sklearn.mixture import BayesianGaussianMixture

from sdgx.models.components.sdv_rdt.transformers.base import BaseTransformer
from sdgx.models.components.sdv_rdt.transformers.null import NullTransformer

EPSILON = np.finfo(np.float32).eps
MAX_DECIMALS = sys.float_info.dig - 1
INTEGER_BOUNDS = {
    "Int8": (-(2**7), 2**7 - 1),
    "Int16": (-(2**15), 2**15 - 1),
    "Int32": (-(2**31), 2**31 - 1),
    "Int64": (-(2**63), 2**63 - 1),
    "UInt8": (0, 2**8 - 1),
    "UInt16": (0, 2**16 - 1),
    "UInt32": (0, 2**32 - 1),
    "UInt64": (0, 2**64 - 1),
}


class FloatFormatter(BaseTransformer):
    """Transformer for numerical data.

    This transformer replaces integer values with their float equivalent.
    Non null float values are not modified.

    Null values are replaced using a ``NullTransformer``.

    Args:
        missing_value_replacement (object or None):
            Indicate what to do with the null values. If an integer or float is given,
            replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are
            given, replace them with the corresponding aggregation. If ``None`` is given,
            do not replace them. Defaults to ``None``.
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
        learn_rounding_scheme (bool):
            Whether or not to learn what place to round to based on the data seen during ``fit``.
            If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
            Defaults to ``False``.
        enforce_min_max_values (bool):
            Whether or not to clip the data returned by ``reverse_transform`` to the min and
            max values seen during ``fit``. Defaults to ``False``.
        computer_representation (dtype):
            Accepts ``'Int8'``, ``'Int16'``, ``'Int32'``, ``'Int64'``, ``'UInt8'``, ``'UInt16'``,
            ``'UInt32'``, ``'UInt64'``, ``'Float'``.
            Defaults to ``'Float'``.
    """

    INPUT_SDTYPE = "numerical"
    DETERMINISTIC_TRANSFORM = True
    DETERMINISTIC_REVERSE = True
    COMPOSITION_IS_IDENTITY = True

    null_transformer = None
    missing_value_replacement = None
    _dtype = None
    _rounding_digits = None
    _min_value = None
    _max_value = None

    def __init__(
        self,
        missing_value_replacement=None,
        model_missing_values=False,
        learn_rounding_scheme=False,
        enforce_min_max_values=False,
        computer_representation="Float",
    ):
        self.missing_value_replacement = missing_value_replacement
        self.model_missing_values = model_missing_values
        self.learn_rounding_scheme = learn_rounding_scheme
        self.enforce_min_max_values = enforce_min_max_values
        self.computer_representation = computer_representation

    def get_output_sdtypes(self):
        """Return the output sdtypes supported by the transformer.

        Returns:
            dict:
                Mapping from the transformed column names to supported sdtypes.
        """
        output_sdtypes = {
            "value": "float",
        }
        if self.null_transformer and self.null_transformer.models_missing_values():
            output_sdtypes["is_null"] = "float"

        return self._add_prefix(output_sdtypes)

    def is_composition_identity(self):
        """Return whether composition of transform and reverse transform produces the input data.

        Returns:
            bool:
                Whether or not transforming and then reverse transforming returns the input data.
        """
        if self.null_transformer and not self.null_transformer.models_missing_values():
            return False

        return self.COMPOSITION_IS_IDENTITY

    @staticmethod
    def _learn_rounding_digits(data):
        # check if data has any decimals
        data = np.array(data)
        roundable_data = data[~(np.isinf(data) | pd.isna(data))]
        if ((roundable_data % 1) != 0).any():
            if (roundable_data == roundable_data.round(MAX_DECIMALS)).all():
                for decimal in range(MAX_DECIMALS + 1):
                    if (roundable_data == roundable_data.round(decimal)).all():
                        return decimal

        return None

    def _raise_out_of_bounds_error(self, value, name, bound_type, min_bound, max_bound):
        raise ValueError(
            f"The {bound_type} value in column '{name}' is {value}."
            f" All values represented by '{self.computer_representation}'"
            f" must be in the range [{min_bound}, {max_bound}]."
        )

    def _validate_values_within_bounds(self, data):
        if self.computer_representation != "Float":
            fractions = data[~data.isna() & data % 1 != 0]
            if not fractions.empty:
                raise ValueError(
                    f"The column '{data.name}' contains float values {fractions.tolist()}. "
                    f"All values represented by '{self.computer_representation}' must be integers."
                )

            min_value = data.min()
            max_value = data.max()
            min_bound, max_bound = INTEGER_BOUNDS[self.computer_representation]
            if min_value < min_bound:
                self._raise_out_of_bounds_error(
                    min_value, data.name, "minimum", min_bound, max_bound
                )

            if max_value > max_bound:
                self._raise_out_of_bounds_error(
                    max_value, data.name, "maximum", min_bound, max_bound
                )

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit.
        """
        self._validate_values_within_bounds(data)
        self._dtype = data.dtype

        if self.enforce_min_max_values:
            self._min_value = data.min()
            self._max_value = data.max()

        if self.learn_rounding_scheme:
            self._rounding_digits = self._learn_rounding_digits(data)

        self.null_transformer = NullTransformer(
            self.missing_value_replacement, self.model_missing_values
        )
        self.null_transformer.fit(data)

    def _transform(self, data):
        """Transform numerical data.

        Integer values are replaced by their float equivalent. Non null float values
        are left unmodified.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        self._validate_values_within_bounds(data)
        data = data.astype(np.float64)
        return self.null_transformer.transform(data)

    def _reverse_transform(self, data):
        """Convert data back into the original format.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        if self.missing_value_replacement is not None:
            data = self.null_transformer.reverse_transform(data)

        if self.enforce_min_max_values:
            data = data.clip(self._min_value, self._max_value)
        elif self.computer_representation != "Float":
            min_bound, max_bound = INTEGER_BOUNDS[self.computer_representation]
            data = data.clip(min_bound, max_bound)

        is_integer = np.dtype(self._dtype).kind == "i"
        if self.learn_rounding_scheme or is_integer:
            data = data.round(self._rounding_digits or 0)

        if pd.isna(data).any() and is_integer:
            return data

        return data.astype(self._dtype)


class GaussianNormalizer(FloatFormatter):
    r"""Transformer for numerical data based on copulas transformation.

    Transformation consists on bringing the input data to a standard normal space
    by using a combination of *cdf* and *inverse cdf* transformations:

    Given a variable :math:`x`:

    - Find the best possible marginal or use user specified one, :math:`P(x)`.
    - do :math:`u = \phi (x)` where :math:`\phi` is cumulative density function,
      given :math:`P(x)`.
    - do :math:`z = \phi_{N(0,1)}^{-1}(u)`, where :math:`\phi_{N(0,1)}^{-1}` is
      the *inverse cdf* of a *standard normal* distribution.

    The reverse transform will do the inverse of the steps above and go from :math:`z`
    to :math:`u` and then to :math:`x`.

    Args:
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
        learn_rounding_scheme (bool):
            Whether or not to learn what place to round to based on the data seen during ``fit``.
            If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
            Defaults to ``False``.
        enforce_min_max_values (bool):
            Whether or not to clip the data returned by ``reverse_transform`` to the min and
            max values seen during ``fit``. Defaults to ``False``.
        distribution (copulas.univariate.Univariate or str):
            Copulas univariate distribution to use. Defaults to ``truncated_gaussian``.
            Options include:

                * ``gaussian``: Use a Gaussian distribution.
                * ``gamma``: Use a Gamma distribution.
                * ``beta``: Use a Beta distribution.
                * ``student_t``: Use a Student T distribution.
                * ``gussian_kde``: Use a GaussianKDE distribution. This model is non-parametric,
                  so using this will make ``get_parameters`` unusable.
                * ``truncated_gaussian``: Use a Truncated Gaussian distribution.
    """

    _univariate = None
    COMPOSITION_IS_IDENTITY = False

    def __init__(
        self,
        model_missing_values=False,
        learn_rounding_scheme=False,
        enforce_min_max_values=False,
        distribution="truncated_gaussian",
    ):
        super().__init__(
            missing_value_replacement="mean",
            model_missing_values=model_missing_values,
            learn_rounding_scheme=learn_rounding_scheme,
            enforce_min_max_values=enforce_min_max_values,
        )

        self.distribution = distribution  # Distribution initialized by the user

        self._distributions = self._get_distributions()
        if isinstance(distribution, str):
            distribution = self._distributions[distribution]

        self._distribution = distribution

    @staticmethod
    def _get_distributions():
        try:
            from sdgx.models.components.sdv_copulas import (  # pylint: disable=import-outside-toplevel
                univariate,
            )
        except ImportError as error:
            error.msg += (
                "\n\nIt seems like `copulas` is not installed.\n"
                "Please install it using:\n\n    pip install rdt[copulas]"
            )
            raise

        return {
            "gaussian": univariate.GaussianUnivariate,
            "gamma": univariate.GammaUnivariate,
            "beta": univariate.BetaUnivariate,
            "student_t": univariate.StudentTUnivariate,
            "gaussian_kde": univariate.GaussianKDE,
            "truncated_gaussian": univariate.TruncatedGaussian,
        }

    def _get_univariate(self):
        distribution = self._distribution
        if any(isinstance(distribution, dist) for dist in self._distributions.values()):
            return copy.deepcopy(distribution)
        if isinstance(distribution, tuple):
            return distribution[0](**distribution[1])
        if isinstance(distribution, type) and distribution in self._distributions.values():
            return distribution()

        raise TypeError(f"Invalid distribution: {distribution}")

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit to.
        """
        self._univariate = self._get_univariate()

        super()._fit(data)
        data = super()._transform(data)
        if data.ndim > 1:
            data = data[:, 0]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._univariate.fit(data)

    def _copula_transform(self, data):
        cdf = self._univariate.cdf(data)
        return scipy.stats.norm.ppf(cdf.clip(0 + EPSILON, 1 - EPSILON))

    def _transform(self, data):
        """Transform numerical data.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        transformed = super()._transform(data)
        if transformed.ndim > 1:
            transformed[:, 0] = self._copula_transform(transformed[:, 0])
        else:
            transformed = self._copula_transform(transformed)

        return transformed

    def _reverse_transform(self, data):
        """Convert data back into the original format.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to transform.

        Returns:
            pandas.Series
        """
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        if data.ndim > 1:
            data[:, 0] = self._univariate.ppf(scipy.stats.norm.cdf(data[:, 0]))
        else:
            data = self._univariate.ppf(scipy.stats.norm.cdf(data))

        return super()._reverse_transform(data)


class ClusterBasedNormalizer(FloatFormatter):
    """Transformer for numerical data using a Bayesian Gaussian Mixture Model.

    This transformation takes a numerical value and transforms it using a Bayesian GMM
    model. It generates two outputs, a discrete value which indicates the selected
    'component' of the GMM and a continuous value which represents the normalized value
    based on the mean and std of the selected component.

    Args:
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
        learn_rounding_scheme (bool):
            Whether or not to learn what place to round to based on the data seen during ``fit``.
            If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
            Defaults to ``False``.
        enforce_min_max_values (bool):
            Whether or not to clip the data returned by ``reverse_transform`` to the min and
            max values seen during ``fit``. Defaults to ``False``.
        max_clusters (int):
            The maximum number of mixture components. Depending on the data, the model may select
            fewer components (based on the ``weight_threshold``).
            Defaults to 10.
        weight_threshold (int, float):
            The minimum value a component weight can take to be considered a valid component.
            ``weights_`` under this value will be ignored.
            Defaults to 0.005.

    Attributes:
        _bgm_transformer:
            An instance of sklearn`s ``BayesianGaussianMixture`` class.
        valid_component_indicator:
            An array indicating the valid components. If the weight of a component is greater
            than the ``weight_threshold``, it's indicated with True, otherwise it's set to False.
    """

    STD_MULTIPLIER = 4
    DETERMINISTIC_TRANSFORM = False
    DETERMINISTIC_REVERSE = True
    COMPOSITION_IS_IDENTITY = False

    _bgm_transformer = None
    valid_component_indicator = None

    def __init__(
        self,
        model_missing_values=False,
        learn_rounding_scheme=False,
        enforce_min_max_values=False,
        max_clusters=10,
        weight_threshold=0.005,
    ):
        super().__init__(
            missing_value_replacement="mean",
            model_missing_values=model_missing_values,
            learn_rounding_scheme=learn_rounding_scheme,
            enforce_min_max_values=enforce_min_max_values,
        )
        self.max_clusters = max_clusters
        self.weight_threshold = weight_threshold

    def get_output_sdtypes(self):
        """Return the output sdtypes supported by the transformer.

        Returns:
            dict:
                Mapping from the transformed column names to supported sdtypes.
        """
        output_sdtypes = {"normalized": "float", "component": "categorical"}
        if self.null_transformer and self.null_transformer.models_missing_values():
            output_sdtypes["is_null"] = "float"

        return self._add_prefix(output_sdtypes)

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit to.
        """
        self._bgm_transformer = BayesianGaussianMixture(
            n_components=self.max_clusters,
            weight_concentration_prior_type="dirichlet_process",
            weight_concentration_prior=0.001,
            n_init=1,
        )

        super()._fit(data)
        data = super()._transform(data)
        if data.ndim > 1:
            data = data[:, 0]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._bgm_transformer.fit(data.reshape(-1, 1))

        self.valid_component_indicator = self._bgm_transformer.weights_ > self.weight_threshold

    def _transform(self, data):
        """Transform the numerical data.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            numpy.ndarray.
        """
        data = super()._transform(data)
        if data.ndim > 1:
            data, model_missing_values = data[:, 0], data[:, 1]

        data = data.reshape((len(data), 1))
        means = self._bgm_transformer.means_.reshape((1, self.max_clusters))

        stds = np.sqrt(self._bgm_transformer.covariances_).reshape((1, self.max_clusters))
        normalized_values = (data - means) / (self.STD_MULTIPLIER * stds)
        normalized_values = normalized_values[:, self.valid_component_indicator]
        component_probs = self._bgm_transformer.predict_proba(data)
        component_probs = component_probs[:, self.valid_component_indicator]

        selected_component = np.zeros(len(data), dtype="int")
        for i in range(len(data)):
            component_prob_t = component_probs[i] + 1e-6
            component_prob_t = component_prob_t / component_prob_t.sum()
            selected_component[i] = np.random.choice(
                np.arange(self.valid_component_indicator.sum()), p=component_prob_t
            )

        aranged = np.arange(len(data))
        normalized = normalized_values[aranged, selected_component].reshape([-1, 1])
        normalized = np.clip(normalized, -0.99, 0.99)
        normalized = normalized[:, 0]
        rows = [normalized, selected_component]
        if self.null_transformer and self.null_transformer.models_missing_values():
            rows.append(model_missing_values)

        return np.stack(rows, axis=1)  # noqa: PD013

    def _reverse_transform_helper(self, data):
        normalized = np.clip(data[:, 0], -1, 1)
        means = self._bgm_transformer.means_.reshape([-1])
        stds = np.sqrt(self._bgm_transformer.covariances_).reshape([-1])
        selected_component = data[:, 1].astype(int)

        std_t = stds[self.valid_component_indicator][selected_component]
        mean_t = means[self.valid_component_indicator][selected_component]
        reversed_data = normalized * self.STD_MULTIPLIER * std_t + mean_t

        return reversed_data

    def _reverse_transform(self, data):
        """Convert data back into the original format.

        Args:
            data (pd.DataFrame or numpy.ndarray):
                Data to transform.

        Returns:
            pandas.Series.
        """
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        recovered_data = self._reverse_transform_helper(data)
        if self.null_transformer and self.null_transformer.models_missing_values():
            data = np.stack([recovered_data, data[:, -1]], axis=1)  # noqa: PD013
        else:
            data = recovered_data

        return super()._reverse_transform(data)



"""Tools to generate strings from regular expressions."""

import re
import string

import numpy as np

import sre_parse  # isort:skip


def _literal(character, max_repeat):
    del max_repeat
    return iter([chr(character)]), 1


def _in(options, max_repeat):
    generators = []
    sizes = []
    for option, args in options:
        generator, size = _GENERATORS[option](args, max_repeat)
        generators.append(generator)
        sizes.append(size)

    return (value for generator in generators for value in generator), np.sum(sizes)


def _range(options, max_repeat):
    del max_repeat
    min_value, max_value = options
    max_value += 1
    return (chr(value) for value in range(min_value, max_value)), max_value - min_value


def _any(options, max_repeat):
    del options
    del max_repeat
    return iter(string.printable), len(string.printable)


def _max_repeat(options, max_repeat):
    min_, max_, options = options
    if max_ == sre_parse.MAXREPEAT:
        max_ = max_repeat

    option, args = options[0]
    _, size = _GENERATORS[option](args, max_repeat)

    generators = []
    sizes = []
    for repeat in range(min_, max_ + 1):
        if repeat:
            sizes.append(size**repeat)
            repeat_generators = [
                (_GENERATORS[option](args, max_repeat)[0], option, args) for _ in range(repeat)
            ]
            generators.append(_from_generators(repeat_generators, max_repeat))

    return (value for generator in generators for value in generator), np.sum(sizes) + int(
        min_ == 0
    )


def _category_chars(regex):
    return [char for char in string.printable if regex.match(char)]


_CATEGORIES = {
    sre_parse.CATEGORY_SPACE: _category_chars(re.compile(r"\s")),
    sre_parse.CATEGORY_NOT_SPACE: _category_chars(re.compile(r"\S")),
    sre_parse.CATEGORY_DIGIT: _category_chars(re.compile(r"\d")),
    sre_parse.CATEGORY_NOT_DIGIT: _category_chars(re.compile(r"\D")),
    sre_parse.CATEGORY_WORD: _category_chars(re.compile(r"\w")),
    sre_parse.CATEGORY_NOT_WORD: _category_chars(re.compile(r"\W")),
}


def _category(category, max_repeat):
    del max_repeat
    characters = _CATEGORIES[category]
    return iter(characters), len(characters)


_GENERATORS = {
    sre_parse.LITERAL: _literal,
    sre_parse.IN: _in,
    sre_parse.RANGE: _range,
    sre_parse.ANY: _any,
    sre_parse.MAX_REPEAT: _max_repeat,
    sre_parse.CATEGORY: _category,
}


def _from_generators(generators, max_repeat):
    previous = [None] + [next(generator) for generator, _, _ in generators[1:]]

    remaining = True
    while remaining:
        generated = []
        for index, (generator, option, args) in enumerate(generators):
            remaining = True
            try:
                value = next(generator)
                generated.append(value)
                previous[index] = value
                generated.extend(previous[index + 1 :])
                break
            except StopIteration:
                generator = _GENERATORS[option](args, max_repeat)[0]
                generators[index] = generator, option, args
                value = next(generator)
                previous[index] = value
                generated.append(value)
                remaining = False

        if remaining:
            yield "".join(reversed(generated))


def strings_from_regex(regex, max_repeat=16):
    """Generate strings that match the given regular expression.

    The output is a generator that produces regular expressions that match
    the indicated regular expressions alongside an integer indicating the
    total length of the generator.

    WARNING: Subpatterns are currently not supported.

    Args:
        regex (str):
            String representing a valid python regular expression.
        max_repeat (int):
            Maximum number of repetitions to produce when the regular
            expression allows an infinte amount. Defaults to 16.

    Returns:
        tuple:
            * Generator that produces strings that match the given regex.
            * Total length of the generator.
    """
    parsed = sre_parse.parse(regex, flags=sre_parse.SRE_FLAG_UNICODE)
    generators = []
    sizes = []
    for option, args in reversed(parsed):
        if option != sre_parse.AT:
            generator, size = _GENERATORS[option](args, max_repeat)
            generators.append((generator, option, args))
            sizes.append(size)

    return _from_generators(generators, max_repeat), np.prod(sizes)



from __future__ import annotations

import datetime
import random
import re

import pandas as pd
import pytest
from faker import Faker
from pydantic import BaseModel, EmailStr

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.generators.chn_pii import ChnPiiGenerator

fake = Faker(locale="zh_CN")
fake_en = Faker(["en_US"])


@pytest.fixture
def chn_personal_test_df():
    row_cnt = 1000
    today = datetime.datetime.today()
    X = []
    header = [
        "ssn_sfz",
        "chn_name",
        "eng_name",
        "gender",
        "birth_date",
        "age",
        "email",
        "mobile_phone_no",
        "chn_address",
        "postcode",
        "job",
        "company_name",
    ]
    for _ in range(row_cnt):
        each_gender = random.choice(["male", "female"])
        if each_gender == "male":
            each_name = fake.last_name() + fake.name_male()
        else:
            each_name = fake.last_name() + fake.name_female()
        each_eng_name = fake_en.name()
        each_birth_date = fake.date()
        each_age = today.year - int(each_birth_date[:4])
        each_email = fake.email()
        each_phone = fake.phone_number()
        each_sfz = fake.ssn()
        each_address = fake.address()
        each_job = fake.job()
        each_corp = fake.company()
        each_postcode = fake.postcode()

        each_x = [
            each_sfz,
            each_name,
            each_eng_name,
            each_gender,
            each_birth_date,
            each_age,
            each_email,
            each_phone,
            each_address,
            each_postcode,
            each_job,
            each_corp,
        ]

        X.append(each_x)

    yield pd.DataFrame(X, columns=header)


def test_chn_pii_generator(chn_personal_test_df: pd.DataFrame):

    assert "chn_name" in chn_personal_test_df.columns
    assert "mobile_phone_no" in chn_personal_test_df.columns
    assert "ssn_sfz" in chn_personal_test_df.columns
    assert "company_name" in chn_personal_test_df.columns

    # get metadata
    metadata_df = Metadata.from_dataframe(chn_personal_test_df)

    # generator
    pii_generator = ChnPiiGenerator()
    assert not pii_generator.fitted
    pii_generator.fit(metadata_df)
    assert pii_generator.fitted
    assert pii_generator.chn_name_columns_list == ["chn_name"]
    assert pii_generator.chn_phone_columns_list == ["mobile_phone_no"]
    assert pii_generator.chn_id_columns_list == ["ssn_sfz"]
    assert pii_generator.chn_company_name_list == ["company_name"]

    converted_df = pii_generator.convert(chn_personal_test_df)
    assert len(converted_df) == len(chn_personal_test_df)
    assert converted_df.shape[1] != chn_personal_test_df.shape[1]
    assert converted_df.shape[1] == chn_personal_test_df.shape[1] - len(
        pii_generator.chn_pii_columns
    )
    assert "chn_name" not in converted_df.columns
    assert "mobile_phone_no" not in converted_df.columns
    assert "ssn_sfz" not in converted_df.columns
    assert "company_name" not in converted_df.columns

    reverse_converted_df = pii_generator.reverse_convert(converted_df)
    assert len(reverse_converted_df) == len(chn_personal_test_df)
    assert "chn_name" in reverse_converted_df.columns
    assert "mobile_phone_no" in reverse_converted_df.columns
    assert "ssn_sfz" in reverse_converted_df.columns
    assert "company_name" in reverse_converted_df.columns
    # each generated value is sfz
    for each_value in chn_personal_test_df["ssn_sfz"].values:
        assert len(each_value) == 18
        pattern = r"^\d{17}[0-9X]$"
        assert bool(re.match(pattern, each_value))
    for each_value in chn_personal_test_df["chn_name"].values:
        pattern = r"^[\u4e00-\u9fa5]{2,5}$"
        assert len(each_value) >= 2 and len(each_value) <= 5
        assert bool(re.match(pattern, each_value))
    for each_value in chn_personal_test_df["mobile_phone_no"].values:
        assert each_value.startswith("1")
        assert len(each_value) == 11
        pattern = r"^1[3-9]\d{9}$"
        assert bool(re.match(pattern, each_value))
    for each_value in chn_personal_test_df["company_name"].values:
        pattern = r".*?公司.*?"
        assert bool(re.match(pattern, each_value))



"""Demo module."""

import pandas as pd

DEMO_URL = "http://ctgan-data.s3.amazonaws.com/census.csv.gz"


def load_demo():
    """Load the demo."""
    return pd.read_csv(DEMO_URL, compression="gzip")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pandas as pd

if TYPE_CHECKING:
    from sdgx.data_models.metadata import Metadata

from sdgx.data_models.relationship import Relationship
from sdgx.exceptions import InspectorInitError


class Inspector:
    """
    Base Inspector class

    Inspector is used to inspect data and generate metadata automatically.

    Parameters:
        ready (bool): Ready to inspect, maybe all fields are fitted, or indicate if there is more data, inspector will be more precise.
    """

    pii = False
    """
    PII refers if a column contains private or sensitive information.
    """

    _inspect_level: int = 10
    """
    Private variable used to store property inspect_level's value.
    """

    ready: bool = False
    """
    Indicates whether the inspector has completed its inference.

    When completed, ready == True.
    """

    @property
    def inspect_level(self):
        """
        Inspected level is a concept newly introduced in version 0.1.6. Since a single column in the table may be marked by different inspectors at the same time (for example: the email column may be recognized as email, but it may also be recognized as the id column, and it may also be recognized by different inspectors at the same time identified as a discrete column, which will cause confusion in subsequent processing), the inspect_leve is used when determining the specific type of a column.

        We will preset different inspector levels for different inspectors, usually more specific inspectors will get higher levels, and general inspectors (like discrete) will have inspect_level.

        The value of the variable inspect_level is limited to 1-100. In baseclass and bool, discrete and numeric types, the inspect_level is set to 10. For datetime and id types, the inspect_level is set to 20.

        Current inspect_level value will make it easier for developers to insert a custom inspector from the middle.
        """
        return self._inspect_level

    @inspect_level.setter
    def inspect_level(self, value: int):
        if value > 0 and value <= 100:
            self._inspect_level = value
        else:
            raise InspectorInitError("The inspect_level should be set in [1, 100].")

    def __init__(self, inspect_level=None, *args, **kwargs):
        self.ready: bool = False
        # add inspect_level check
        if inspect_level:
            self.inspect_level = inspect_level

    def fit(self, raw_data: pd.DataFrame, *args, **kwargs):
        """Fit the inspector.

        Args:
            raw_data (pd.DataFrame): Raw data
        """
        return

    def inspect(self, *args, **kwargs) -> dict[str, Any]:
        """Inspect raw data and generate metadata."""


class RelationshipInspector(Inspector):
    """
    Empty RelationshipInspector for inheritence

    Subclass should implement `_build_relationship` and `fit`
    """

    def _build_relationship(self) -> list[Relationship]:
        return []

    def fit(
        self,
        raw_data: pd.DataFrame,
        name: str | None = None,
        metadata: "Metadata" | None = None,
        *args,
        **kwargs,
    ):
        pass

    def inspect(self, *args, **kwargs) -> dict[str, Any]:
        """Inspect raw data and generate metadata."""
        return {"relationships": self._build_relationship()}



from __future__ import annotations

import math
import time
from pathlib import Path
from typing import List, Optional

import numpy as np
import pandas as pd
import torch
from packaging import version
from torch import optim
from torch.nn import (
    BatchNorm1d,
    Dropout,
    LeakyReLU,
    Linear,
    Module,
    ReLU,
    Sequential,
    functional,
)
from tqdm import autonotebook as tqdm

from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.models.components.optimize.ndarray_loader import NDArrayLoader
from sdgx.models.components.optimize.sdv_ctgan.data_sampler import DataSampler
from sdgx.models.components.optimize.sdv_ctgan.data_transformer import DataTransformer
from sdgx.models.components.sdv_ctgan.synthesizers.base import (
    BaseSynthesizer as SDVBaseSynthesizer,
)
from sdgx.models.components.sdv_ctgan.synthesizers.base import (
    BatchedSynthesizer,
    random_state,
)
from sdgx.models.extension import hookimpl
from sdgx.models.ml.single_table.base import MLSynthesizerModel
from sdgx.utils import logger


class Discriminator(Module):
    """Discriminator for the CTGAN."""

    def __init__(self, input_dim, discriminator_dim, pac=10):
        super(Discriminator, self).__init__()
        dim = input_dim * pac
        self.pac = pac
        self.pacdim = dim
        seq = []
        for item in list(discriminator_dim):
            seq += [Linear(dim, item), LeakyReLU(0.2), Dropout(0.5)]
            dim = item

        seq += [Linear(dim, 1)]
        self.seq = Sequential(*seq)

    def calc_gradient_penalty(self, real_data, fake_data, device="cpu", pac=10, lambda_=10):
        """Compute the gradient penalty."""
        alpha = torch.rand(real_data.size(0) // pac, 1, 1, device=device)
        alpha = alpha.repeat(1, pac, real_data.size(1))
        alpha = alpha.view(-1, real_data.size(1))

        interpolates = alpha * real_data + ((1 - alpha) * fake_data)

        disc_interpolates = self(interpolates)

        gradients = torch.autograd.grad(
            outputs=disc_interpolates,
            inputs=interpolates,
            grad_outputs=torch.ones(disc_interpolates.size(), device=device),
            create_graph=True,
            retain_graph=True,
            only_inputs=True,
        )[0]

        gradients_view = gradients.view(-1, pac * real_data.size(1)).norm(2, dim=1) - 1
        gradient_penalty = ((gradients_view) ** 2).mean() * lambda_

        return gradient_penalty

    def forward(self, input_):
        """Apply the Discriminator to the `input_`."""
        assert input_.size()[0] % self.pac == 0
        return self.seq(input_.view(-1, self.pacdim))


class Residual(Module):
    """Residual layer for the CTGAN."""

    def __init__(self, i, o):
        super(Residual, self).__init__()
        self.fc = Linear(i, o)
        self.bn = BatchNorm1d(o)
        self.relu = ReLU()

    def forward(self, input_):
        """Apply the Residual layer to the `input_`."""
        out = self.fc(input_)
        out = self.bn(out)
        out = self.relu(out)
        return torch.cat([out, input_], dim=1)


class Generator(Module):
    """Generator for the CTGAN."""

    def __init__(self, embedding_dim, generator_dim, data_dim):
        super(Generator, self).__init__()
        dim = embedding_dim
        seq = []
        for item in list(generator_dim):
            seq += [Residual(dim, item)]
            dim += item
        seq.append(Linear(dim, data_dim))
        self.seq = Sequential(*seq)

    def forward(self, input_):
        """Apply the Generator to the `input_`."""
        data = self.seq(input_)
        return data


class CTGANSynthesizerModel(MLSynthesizerModel, BatchedSynthesizer):
    """
    Modified from ``sdgx.models.components.sdv_ctgan.synthesizers.ctgan.CTGANSynthesizer``.
    A CTGANSynthesizer but provided :ref:`SynthesizerModel` interface with chunked fit.

    This is the core class of the CTGAN project, where the different components
    are orchestrated together.
    For more details about the process, please check the [Modeling Tabular data using
    Conditional GAN](https://arxiv.org/abs/1907.00503) paper.


    Args:
        embedding_dim (int):
            Size of the random sample passed to the Generator. Defaults to 128.
        generator_dim (tuple or list of ints):
            Size of the output samples for each one of the Residuals. A Residual Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        discriminator_dim (tuple or list of ints):
            Size of the output samples for each one of the Discriminator Layers. A Linear Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        generator_lr (float):
            Learning rate for the generator. Defaults to 2e-4.
        generator_decay (float):
            Generator weight decay for the Adam Optimizer. Defaults to 1e-6.
        discriminator_lr (float):
            Learning rate for the discriminator. Defaults to 2e-4.
        discriminator_decay (float):
            Discriminator weight decay for the Adam Optimizer. Defaults to 1e-6.
        batch_size (int):
            Number of data samples to process in each step.
        discriminator_steps (int):
            Number of discriminator updates to do for each generator update.
            From the WGAN paper: https://arxiv.org/abs/1701.07875. WGAN paper
            default is 5. Default used is 1 to match original CTGAN implementation.
        log_frequency (boolean):
            Whether to use log frequency of categorical levels in conditional
            sampling. Defaults to ``True``.
        epochs (int):
            Number of training epochs. Defaults to 300.
        pac (int):
            Number of samples to group together when applying the discriminator.
            Defaults to 10.
        device (str):
            Device to run the training on. Preferred to be 'cuda' for GPU if available.
    """

    MODEL_SAVE_NAME = "ctgan.pkl"

    def __init__(
        self,
        embedding_dim=128,
        generator_dim=(256, 256),
        discriminator_dim=(256, 256),
        generator_lr=2e-4,
        generator_decay=1e-6,
        discriminator_lr=2e-4,
        discriminator_decay=1e-6,
        batch_size=500,
        discriminator_steps=1,
        log_frequency=True,
        epochs=300,
        pac=10,
        device="cuda" if torch.cuda.is_available() else "cpu",
    ):
        assert batch_size % 2 == 0
        BatchedSynthesizer.__init__(self, batch_size=batch_size)
        self._embedding_dim = embedding_dim
        self._generator_dim = generator_dim
        self._discriminator_dim = discriminator_dim

        self._generator_lr = generator_lr
        self._generator_decay = generator_decay
        self._discriminator_lr = discriminator_lr
        self._discriminator_decay = discriminator_decay

        self._discriminator_steps = discriminator_steps
        self._log_frequency = log_frequency
        self._epochs = epochs
        self.pac = pac

        self._device = torch.device(device)

        # Following components are initialized in `_pre_fit`
        self._transformer: Optional[DataTransformer] = None
        self._data_sampler: Optional[DataSampler] = None
        self._generator = None
        self._ndarry_loader: Optional[NDArrayLoader] = None
        self.data_dim: Optional[int] = None

    def fit(self, metadata: Metadata, dataloader: DataLoader, epochs=None, *args, **kwargs):
        # In the future, sdgx use `sdgx.data_processor.transformers.discrete` to handle discrete_columns
        # the original sdv transformer will be removed in version 0.3.0
        # This will be done in another PR.
        discrete_columns = list(metadata.get("discrete_columns"))
        if epochs is not None:
            self._epochs = epochs
        self._pre_fit(dataloader, discrete_columns, metadata)
        if self.fit_data_empty:
            logger.info("CTGAN fit finished because of empty df detected.")
            return
        logger.info("CTGAN prefit finished, start CTGAN training.")
        self._fit(len(self._ndarry_loader))
        logger.info("CTGAN training finished.")

    def _pre_fit(
        self, dataloader: DataLoader, discrete_columns: list[str] = None, metadata: Metadata = None
    ):
        if not discrete_columns:
            discrete_columns = []

        # self._validate_discrete_columns(dataloader.columns(), discrete_columns)
        discrete_columns = self._filter_discrete_columns(dataloader.columns(), discrete_columns)
        # if the df is empty, we don't need to do anything
        if self.fit_data_empty:
            return
        # Fit Transformer and DataSampler
        self._transformer = DataTransformer(metadata=metadata)
        logger.info("Fitting model's transformer...")
        self._transformer.fit(dataloader, discrete_columns)
        logger.info("Transforming data...")
        self._ndarry_loader = self._transformer.transform(dataloader)

        logger.info("Sampling data.")
        self._data_sampler = DataSampler(
            self._ndarry_loader, self._transformer.output_info_list, self._log_frequency
        )

        logger.info("Initialize Generator.")
        # Initialize Generator
        self.data_dim = self._transformer.output_dimensions
        self._generator = Generator(
            self._embedding_dim + self._data_sampler.dim_cond_vec(),
            self._generator_dim,
            self.data_dim,
        ).to(self._device)

    @random_state
    def _fit(self, data_size: int):
        """Fit the CTGAN Synthesizer models to the training data."""
        logger.info(f"Fit using data_size:{data_size}, data_dim: {self.data_dim}.")
        epochs = self._epochs
        # data_dim = self._transformer.output_dimensions
        discriminator = Discriminator(
            self.data_dim + self._data_sampler.dim_cond_vec(),
            self._discriminator_dim,
            pac=self.pac,
        ).to(self._device)

        optimizerG = optim.Adam(
            self._generator.parameters(),
            lr=self._generator_lr,
            betas=(0.5, 0.9),
            weight_decay=self._generator_decay,
        )

        optimizerD = optim.Adam(
            discriminator.parameters(),
            lr=self._discriminator_lr,
            betas=(0.5, 0.9),
            weight_decay=self._discriminator_decay,
        )

        mean = torch.zeros(self._batch_size, self._embedding_dim, device=self._device)
        std = mean + 1

        logger.info("Starting model training, epochs: {}".format(epochs))
        steps_per_epoch = max(data_size // self._batch_size, 1)
        for i in range(epochs):
            start_time = time.time()
            for id_ in tqdm.tqdm(range(steps_per_epoch), desc="Fitting batches", delay=3):
                for n in range(self._discriminator_steps):
                    fakez = torch.normal(mean=mean, std=std)

                    condvec = self._data_sampler.sample_condvec(self._batch_size)
                    if condvec is None:
                        c1, m1, col, opt = None, None, None, None
                        real = self._data_sampler.sample_data(self._batch_size, col, opt)
                    else:
                        c1, m1, col, opt = condvec
                        c1 = torch.from_numpy(c1).to(self._device)
                        m1 = torch.from_numpy(m1).to(self._device)
                        fakez = torch.cat([fakez, c1], dim=1)

                        perm = np.arange(self._batch_size)
                        np.random.shuffle(perm)
                        real = self._data_sampler.sample_data(
                            self._batch_size, col[perm], opt[perm]
                        )
                        c2 = c1[perm]

                    fake = self._generator(fakez)
                    fakeact = self._apply_activate(fake)

                    real = torch.from_numpy(real.astype("float32")).to(self._device)

                    if c1 is not None:
                        fake_cat = torch.cat([fakeact, c1], dim=1)
                        real_cat = torch.cat([real, c2], dim=1)
                    else:
                        real_cat = real
                        fake_cat = fakeact

                    y_fake = discriminator(fake_cat)
                    y_real = discriminator(real_cat)

                    pen = discriminator.calc_gradient_penalty(
                        real_cat, fake_cat, self._device, self.pac
                    )
                    loss_d = -(torch.mean(y_real) - torch.mean(y_fake))

                    optimizerD.zero_grad()
                    pen.backward(retain_graph=True)
                    loss_d.backward()
                    optimizerD.step()

                fakez = torch.normal(mean=mean, std=std)
                condvec = self._data_sampler.sample_condvec(self._batch_size)

                if condvec is None:
                    c1, m1, col, opt = None, None, None, None
                else:
                    c1, m1, col, opt = condvec
                    c1 = torch.from_numpy(c1).to(self._device)
                    m1 = torch.from_numpy(m1).to(self._device)
                    fakez = torch.cat([fakez, c1], dim=1)

                fake = self._generator(fakez)
                fakeact = self._apply_activate(fake)

                if c1 is not None:
                    y_fake = discriminator(torch.cat([fakeact, c1], dim=1))
                else:
                    y_fake = discriminator(fakeact)

                if condvec is None:
                    cross_entropy = 0
                else:
                    cross_entropy = self._cond_loss(fake, c1, m1)

                loss_g = -torch.mean(y_fake) + cross_entropy

                optimizerG.zero_grad()
                loss_g.backward()
                optimizerG.step()

            logger.info(
                f"Epoch {i+1}, Loss G: {loss_g.detach().cpu(): .4f},"  # noqa: T001
                f" Loss D: {loss_d.detach().cpu(): .4f},"
                f" Time: {time.time() - start_time: .4f}",
            )

    def sample(self, count: int, *args, **kwargs) -> pd.DataFrame:
        if self.fit_data_empty:
            return pd.DataFrame(index=range(count))
        return self._sample(count, *args, **kwargs)

    @random_state
    def _sample(self, n, condition_column=None, condition_value=None, drop_more=True):
        """Sample data similar to the training data.

        Choosing a condition_column and condition_value will increase the probability of the
        discrete condition_value happening in the condition_column.

        Args:
            n (int):
                Number of rows to sample.
            condition_column (string):
                Name of a discrete column.
            condition_value (string):
                Name of the category in the condition_column which we wish to increase the
                probability of happening.

        Returns:
            numpy.ndarray or pandas.DataFrame
        """
        if condition_column is not None and condition_value is not None:
            condition_info = self._transformer.convert_column_name_value_to_id(
                condition_column, condition_value
            )
            global_condition_vec = self._data_sampler.generate_cond_from_condition_column_info(
                condition_info, self._batch_size
            )
        else:
            global_condition_vec = None

        steps = math.ceil(n / self._batch_size)
        data = []
        for _ in tqdm.tqdm(range(steps), desc="Sampling batches", delay=3):
            mean = torch.zeros(self._batch_size, self._embedding_dim)
            std = mean + 1
            fakez = torch.normal(mean=mean, std=std).to(self._device)

            if global_condition_vec is not None:
                condvec = global_condition_vec.copy()
            else:
                condvec = self._data_sampler.sample_original_condvec(self._batch_size)

            if condvec is None:
                pass
            else:
                c1 = condvec
                c1 = torch.from_numpy(c1).to(self._device)
                fakez = torch.cat([fakez, c1], dim=1)

            fake = self._generator(fakez)
            fakeact = self._apply_activate(fake)
            data.append(fakeact.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)
        logger.info("CTGAN Generated {} raw samples.".format(data.shape[0]))
        if drop_more:
            data = data[:n]
        return self._transformer.inverse_transform(data)

    def save(self, save_dir: str | Path):
        save_dir.mkdir(parents=True, exist_ok=True)
        return SDVBaseSynthesizer.save(self, save_dir / self.MODEL_SAVE_NAME)

    @classmethod
    def load(cls, save_dir: str | Path, device: str = None) -> "CTGANSynthesizerModel":
        return SDVBaseSynthesizer.load(save_dir / cls.MODEL_SAVE_NAME, device)

    @staticmethod
    def _gumbel_softmax(logits, tau=1, hard=False, eps=1e-10, dim=-1):
        """Deals with the instability of the gumbel_softmax for older versions of torch.

        For more details about the issue:
        https://drive.google.com/file/d/1AA5wPfZ1kquaRtVruCd6BiYZGcDeNxyP/view?usp=sharing

        Args:
            logits […, num_features]:
                Unnormalized log probabilities
            tau:
                Non-negative scalar temperature
            hard (bool):
                If True, the returned samples will be discretized as one-hot vectors,
                but will be differentiated as if it is the soft sample in autograd
            dim (int):
                A dimension along which softmax will be computed. Default: -1.

        Returns:
            Sampled tensor of same shape as logits from the Gumbel-Softmax distribution.
        """
        if version.parse(torch.__version__) < version.parse("1.2.0"):
            for i in range(10):
                transformed = functional.gumbel_softmax(
                    logits, tau=tau, hard=hard, eps=eps, dim=dim
                )
                if not torch.isnan(transformed).any():
                    return transformed
            raise ValueError("gumbel_softmax returning NaN.")

        return functional.gumbel_softmax(logits, tau=tau, hard=hard, eps=eps, dim=dim)

    def _apply_activate(self, data):
        """Apply proper activation function to the output of the generator."""
        data_t = []
        st = 0
        for column_info in self._transformer.output_info_list:
            for span_info in column_info:
                if span_info.activation_fn == "tanh":
                    ed = st + span_info.dim
                    data_t.append(torch.tanh(data[:, st:ed]))
                    st = ed
                elif span_info.activation_fn == "softmax":
                    ed = st + span_info.dim
                    transformed = self._gumbel_softmax(data[:, st:ed], tau=0.2)
                    data_t.append(transformed)
                    st = ed
                elif span_info.activation_fn == "linear":
                    # for label encoder
                    ed = st + span_info.dim
                    transformed = data[:, st:ed].clone()
                    data_t.append(transformed)
                    st = ed
                else:
                    raise ValueError(f"Unexpected activation function {span_info.activation_fn}.")

        return torch.cat(data_t, dim=1)

    def _cond_loss(self, data, c, m):
        """Compute the cross entropy loss on the fixed discrete column."""
        loss = []
        st = 0
        st_c = 0
        for column_info in self._transformer.output_info_list:
            for span_info in column_info:
                if len(column_info) != 1 or span_info.activation_fn != "softmax":
                    # not discrete column
                    st += span_info.dim
                else:
                    ed = st + span_info.dim
                    ed_c = st_c + span_info.dim
                    tmp = functional.cross_entropy(
                        data[:, st:ed],
                        torch.argmax(c[:, st_c:ed_c], dim=1),
                        reduction="none",
                    )
                    loss.append(tmp)
                    st = ed
                    st_c = ed_c

        loss = torch.stack(loss, dim=1)  # noqa: PD013

        return (loss * m).sum() / data.size()[0]

    def _filter_discrete_columns(self, train_data: List[str], discrete_columns: List[str]):
        """
        We filter PII Column here, which PII would only be discrete for now.
        As PII would be generating from PII Generator which not synthetic from model.

        Besides we need to figure it out when to stop model fitting:
        The original data consists entirely of discrete column data, and all of this discrete column data is PII.

        For `train_data`, there are three possibilities for the columns type.
         - train_data = valid_discrete + valid_continue
         - train_data = valid_continue
         - train_data = valid_discrete

        For `discrete_columns`, discrete_columns = invalid_discrete(PII) + valid_discrete

        Thus, valid_discrete = discrete_columns - invalid_discrete
                             = discrete_columns - Set.intersection(train_data, discrete_columns)

        Thus, original_data_is_all_PII: discrete_columns is not empty & train_data is empty
        """
        # Discrete_columns is empty - simple an empty list, but we need to continue fitting continue columns
        if len(discrete_columns) == 0:
            return discrete_columns

        # Discrete_columns is not empty - check if train_data is empty for stop model fitting
        if len(train_data) == 0:
            self.fit_data_empty = True
            return discrete_columns

        # Filter valid discrete columns
        invalid_columns = set(discrete_columns) - set(train_data)
        return set(discrete_columns) - set(invalid_columns)

    def _validate_discrete_columns(self, train_data, discrete_columns):
        """Check whether ``discrete_columns`` exists in ``train_data``.

        Args:
            train_data (numpy.ndarray or pandas.DataFrame or list):
                Training Data. It must be a 2-dimensional numpy array or a pandas.DataFrame.
            discrete_columns (list-like):
                List of discrete columns to be used to generate the Conditional
                Vector. If ``train_data`` is a Numpy array, this list should
                contain the integer indices of the columns. Otherwise, if it is
                a ``pandas.DataFrame``, this list should contain the column names.
        """
        if isinstance(train_data, pd.DataFrame):
            invalid_columns = set(discrete_columns) - set(train_data.columns)
        elif isinstance(train_data, np.ndarray):
            invalid_columns = []
            for column in discrete_columns:
                if column < 0 or column >= train_data.shape[1]:
                    invalid_columns.append(column)
        elif isinstance(train_data, list):
            invalid_columns = set(discrete_columns) - set(train_data)
        else:
            raise TypeError("``train_data`` should be either pd.DataFrame or np.array.")

        if invalid_columns:
            raise ValueError(f"Invalid columns found: {invalid_columns}")

    def set_device(self, device):
        """Set the `device` to be used ('GPU' or 'CPU)."""
        self._device = device
        if self._generator is not None:
            self._generator.to(self._device)


@hookimpl
def register(manager):
    manager.register("CTGAN", CTGANSynthesizerModel)



import pytest

from sdgx.data_connectors.manager import DataConnectorManager


@pytest.fixture
def manager():
    yield DataConnectorManager()


@pytest.mark.parametrize(
    "supported_data_connector",
    ["CsvConnector"],
)
def test_manager(supported_data_connector, manager: DataConnectorManager):
    assert manager._normalize_name(supported_data_connector) in manager.registed_data_connectors


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



"""Copulas optimization functions."""

import numpy as np


def bisect(f, xmin, xmax, tol=1e-8, maxiter=50):
    """Bisection method for finding roots.

    This method implements a simple vectorized routine for identifying
    the root (of a monotonically increasing function) given a bracketing
    interval.

    Arguments:
        f (Callable):
            A function which takes as input a vector x and returns a
            vector with the same number of dimensions.
        xmin (np.ndarray):
            The minimum value for x such that f(x) <= 0.
        xmax (np.ndarray):
            The maximum value for x such that f(x) >= 0.

    Returns:
        numpy.ndarray:
            The value of x such that f(x) is close to 0.
    """
    assert (f(xmin) <= 0.0).all()
    assert (f(xmax) >= 0.0).all()

    for _ in range(maxiter):
        guess = (xmin + xmax) / 2.0
        fguess = f(guess)
        xmin[fguess <= 0] = guess[fguess <= 0]
        xmax[fguess >= 0] = guess[fguess >= 0]
        if (xmax - xmin).max() < tol:
            break

    return (xmin + xmax) / 2.0


def chandrupatla(f, xmin, xmax, eps_m=None, eps_a=None, maxiter=50):
    """Chandrupatla's algorithm.

    This is adapted from [1] which implements Chandrupatla's algorithm [2]
    which starts from a bracketing interval and, conditionally, swaps between
    bisection and inverse quadratic interpolation.

    [1] https://github.com/scipy/scipy/issues/7242#issuecomment-290548427
    [2] https://books.google.com/books?id=cC-8BAAAQBAJ&pg=PA95

    Arguments:
        f (Callable):
            A function which takes as input a vector x and returns a
            vector with the same number of dimensions.
        xmin (np.ndarray):
            The minimum value for x such that f(x) <= 0.
        xmax (np.ndarray):
            The maximum value for x such that f(x) >= 0.

    Returns:
        numpy.ndarray:
            The value of x such that f(x) is close to 0.
    """
    # Initialization
    a = xmax
    b = xmin
    fa = f(a)
    fb = f(b)

    # Make sure we know the size of the result
    shape = np.shape(fa)
    assert shape == np.shape(fb)

    fc = fa
    c = a

    # Make sure we are bracketing a root in each case
    assert (np.sign(fa) * np.sign(fb) <= 0).all()
    t = 0.5
    # Initialize an array of False,
    # determines whether we should do inverse quadratic interpolation
    iqi = np.zeros(shape, dtype=bool)

    # jms: some guesses for default values of the eps_m and eps_a settings
    # based on machine precision... not sure exactly what to do here
    eps = np.finfo(float).eps
    if eps_m is None:
        eps_m = eps
    if eps_a is None:
        eps_a = 2 * eps

    iterations = 0
    terminate = False

    while maxiter > 0:
        maxiter -= 1
        # use t to linearly interpolate between a and b,
        # and evaluate this function as our newest estimate xt
        xt = np.clip(a + t * (b - a), xmin, xmax)
        ft = f(xt)

        # update our history of the last few points so that
        # - a is the newest estimate (we're going to update it from xt)
        # - c and b get the preceding two estimates
        # - a and b maintain opposite signs for f(a) and f(b)
        samesign = np.sign(ft) == np.sign(fa)
        c = np.choose(samesign, [b, a])
        b = np.choose(samesign, [a, b])
        fc = np.choose(samesign, [fb, fa])
        fb = np.choose(samesign, [fa, fb])
        a = xt
        fa = ft

        # set xm so that f(xm) is the minimum magnitude of f(a) and f(b)
        fa_is_smaller = np.abs(fa) < np.abs(fb)
        xm = np.choose(fa_is_smaller, [b, a])
        fm = np.choose(fa_is_smaller, [fb, fa])

        tol = 2 * eps_m * np.abs(xm) + eps_a
        tlim = tol / np.abs(b - c)
        terminate = np.logical_or(terminate, np.logical_or(fm == 0, tlim > 0.5))

        if np.all(terminate):
            break
        iterations += 1 - terminate

        # Figure out values xi and phi
        # to determine which method we should use next
        xi = (a - b) / (c - b)
        phi = (fa - fb) / (fc - fb)
        iqi = np.logical_and(phi**2 < xi, (1 - phi) ** 2 < 1 - xi)

        if not shape:
            # scalar case
            if iqi:
                # inverse quadratic interpolation
                eq1 = fa / (fb - fa) * fc / (fb - fc)
                eq2 = (c - a) / (b - a) * fa / (fc - fa) * fb / (fc - fb)
                t = eq1 + eq2
            else:
                # bisection
                t = 0.5
        else:
            # array case
            t = np.full(shape, 0.5)
            a2, b2, c2, fa2, fb2, fc2 = a[iqi], b[iqi], c[iqi], fa[iqi], fb[iqi], fc[iqi]
            t[iqi] = fa2 / (fb2 - fa2) * fc2 / (fb2 - fc2) + (c2 - a2) / (b2 - a2) * fa2 / (
                fc2 - fa2
            ) * fb2 / (fc2 - fb2)

        # limit to the range (tlim, 1-tlim)
        t = np.minimum(1 - tlim, np.maximum(tlim, t))

    # done!
    return xm



"""Multivariate copulas module."""

from sdgx.models.components.sdv_copulas.multivariate.base import Multivariate
from sdgx.models.components.sdv_copulas.multivariate.gaussian import (
    GaussianMultivariate,
)
from sdgx.models.components.sdv_copulas.multivariate.tree import Tree, TreeTypes
from sdgx.models.components.sdv_copulas.multivariate.vine import VineCopula

__all__ = ("Multivariate", "GaussianMultivariate", "VineCopula", "Tree", "TreeTypes")



from __future__ import annotations

import copy
from typing import Any

import pandas as pd

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.transformers.base import Transformer
from sdgx.utils import logger


class ConstValueTransformer(Transformer):
    """
    A transformer that replaces the input with a constant value.

    This class is used to transform any input data into a predefined constant value.
    It is particularly useful in scenarios where a consistent output is required regardless of the input.

    Attributes:
        const_value (dict[Any]): The constant value that will be returned.
    """

    const_columns: list

    const_values: dict[Any, Any]

    def __init__(self):
        self.const_columns = []
        self.const_values = {}

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):
        """
        Fit method for the transformer.

        This method processes the metadata to identify columns that should be replaced with a constant value.
        It updates the internal state of the transformer with the columns and their corresponding constant values.

        Args:
            metadata (Metadata | None): The metadata object containing information about the columns and their data types.
            **kwargs (dict[str, Any]): Additional keyword arguments.

        Returns:
            None
        """

        for each_col in metadata.column_list:
            if metadata.get_column_data_type(each_col) == "const":
                self.const_columns.append(each_col)
                # self.const_values[each_col] = metadata.get("const_values")[each_col]

        logger.info("ConstValueTransformer Fitted.")

        self.fitted = True

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        Convert method to handle missing values in the input data by replacing specified columns with constant values.

        This method iterates over the columns identified for replacement with constant values and removes them from the input DataFrame.
        The removal is based on the columns specified during the fitting process.

        Args:
            raw_data (pd.DataFrame): The input DataFrame containing the data to be processed.

        Returns:
            pd.DataFrame: A DataFrame with the specified columns removed.
        """

        processed_data = copy.deepcopy(raw_data)

        logger.info("Converting data using ConstValueTransformer...")

        for each_col in self.const_columns:
            # record values here
            if each_col not in self.const_values.keys():
                self.const_values[each_col] = processed_data[each_col].unique()[0]
            processed_data = self.remove_columns(processed_data, [each_col])

        logger.info("Converting data using ConstValueTransformer... Finished.")

        return processed_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        Reverse_convert method for the transformer.

        This method restores the original columns that were replaced with constant values during the conversion process.
        It iterates over the columns identified for replacement with constant values and adds them back to the DataFrame
        with the predefined constant values.

        Args:
            processed_data (pd.DataFrame): The input DataFrame containing the processed data.

        Returns:
            pd.DataFrame: A DataFrame with the original columns restored, filled with their corresponding constant values.
        """
        df_length = processed_data.shape[0]

        for each_col_name in self.const_columns:
            each_value = self.const_values[each_col_name]
            each_const_col = [each_value for _ in range(df_length)]
            each_const_df = pd.DataFrame({each_col_name: each_const_col})
            processed_data = self.attach_columns(processed_data, each_const_df)

        logger.info("Data reverse-converted by ConstValueTransformer.")

        return processed_data


@hookimpl
def register(manager):
    manager.register("ConstValueTransformer", ConstValueTransformer)



"""TruncatedGaussian module."""

import numpy as np
from scipy.optimize import fmin_slsqp
from scipy.stats import truncnorm

from sdgx.models.components.sdv_copulas import (
    EPSILON,
    store_args,
    validate_random_state,
)
from sdgx.models.components.sdv_copulas.univariate.base import (
    BoundedType,
    ParametricType,
    ScipyModel,
)


class TruncatedGaussian(ScipyModel):
    """Wrapper around scipy.stats.truncnorm.

    Documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.truncnorm.html
    """

    PARAMETRIC = ParametricType.PARAMETRIC
    BOUNDED = BoundedType.BOUNDED
    MODEL_CLASS = truncnorm

    @store_args
    def __init__(self, minimum=None, maximum=None, random_state=None):
        self.random_state = validate_random_state(random_state)
        self.min = minimum
        self.max = maximum

    def _fit_constant(self, X):
        constant = np.unique(X)[0]
        self._params = {"a": constant, "b": constant, "loc": constant, "scale": 0.0}

    def _fit(self, X):
        if self.min is None:
            self.min = X.min() - EPSILON

        if self.max is None:
            self.max = X.max() + EPSILON

        def nnlf(params):
            loc, scale = params
            a = (self.min - loc) / scale
            b = (self.max - loc) / scale
            return truncnorm.nnlf((a, b, loc, scale), X)

        initial_params = X.mean(), X.std()
        optimal = fmin_slsqp(
            nnlf,
            initial_params,
            iprint=False,
            bounds=[(self.min, self.max), (0.0, (self.max - self.min) ** 2)],
        )

        loc, scale = optimal
        a = (self.min - loc) / scale
        b = (self.max - loc) / scale

        self._params = {"a": a, "b": b, "loc": loc, "scale": scale}

    def _is_constant(self):
        return self._params["a"] == self._params["b"]

    def _extract_constant(self):
        return self._params["loc"]



# -*- coding: utf-8 -*-

"""Top-level package for RDT."""


__author__ = """MIT Data To AI Lab"""
__email__ = "dailabmit@gmail.com"
__version__ = "1.2.1"

import numpy as np
import pandas as pd

from sdgx.models.components.sdv_rdt import transformers
from sdgx.models.components.sdv_rdt.hyper_transformer import HyperTransformer

__all__ = ["HyperTransformer", "transformers"]

RANDOM_SEED = 42


def get_demo(num_rows=5):
    """Generate demo data with multiple sdtypes.

    The first five rows are hard coded. The rest are randomly generated
    using ``np.random.seed(42)``.

    Args:
        num_rows (int):
            Number of data rows to generate. Defaults to 5.

    Returns:
        pd.DataFrame
    """
    # Hard code first five rows
    login_dates = ["2021-06-26", "2021-02-10", "NAT", "2020-09-26", "2020-12-22"]
    last_login = [np.datetime64(i) for i in login_dates]
    email_optin = pd.Series([False, False, False, True, np.nan], dtype="object")
    credit_card = ["VISA", "VISA", "AMEX", np.nan, "DISCOVER"]
    age = [29, 18, 21, 45, 32]
    dollars_spent = [99.99, np.nan, 2.50, 25.00, 19.99]

    data = pd.DataFrame(
        {
            "last_login": last_login,
            "email_optin": email_optin,
            "credit_card": credit_card,
            "age": age,
            "dollars_spent": dollars_spent,
        }
    )

    if num_rows <= 5:
        return data.iloc[:num_rows]

    # Randomly generate the remaining rows
    random_state = np.random.get_state()
    np.random.set_state(np.random.RandomState(RANDOM_SEED).get_state())
    try:
        num_rows -= 5

        login_dates = np.array(
            [
                np.datetime64("2000-01-01") + np.timedelta64(np.random.randint(0, 10000), "D")
                for _ in range(num_rows)
            ]
        )
        login_dates[np.random.random(size=num_rows) > 0.8] = np.datetime64("NaT")

        email_optin = pd.Series([True, False, np.nan], dtype="object").sample(
            num_rows, replace=True
        )
        credit_card = np.random.choice(["VISA", "AMEX", np.nan, "DISCOVER"], size=num_rows)
        age = np.random.randint(18, 100, size=num_rows)

        dollars_spent = np.around(np.random.uniform(0, 100, size=num_rows), decimals=2)
        dollars_spent[np.random.random(size=num_rows) > 0.8] = np.nan

    finally:
        np.random.set_state(random_state)

    return data.append(
        pd.DataFrame(
            {
                "last_login": login_dates,
                "email_optin": email_optin,
                "credit_card": credit_card,
                "age": age,
                "dollars_spent": dollars_spent,
            }
        ),
        ignore_index=True,
    )



import pytest

from sdgx.models.components.utils import StrValuedBaseEnum as SE


class A(SE):
    a = "1"
    b = "2"


def test_se():
    a = A("1")
    b = A("2")
    assert isinstance(a.value, str) and a.value == "1"
    assert b.value == "2"
    assert A.values == {"1", "2"}
    assert a != b
    assert ["1", "2", "3"] not in A
    assert "1" in A
    assert ["2"] in A
    assert ["1", "2"] in A
    assert "1" == a
    assert 1 != a
